class Common(object):
  XIAOMI_HEADER_PREFIX = "x-xiaomi-"
  XIAOMI_HEADER_DATE = "x-xiaomi-date"

  GALAXY_ACCESS_KEY_ID = "GalaxyAccessKeyId"
  SIGNATURE = "Signature"
  EXPIRES = "Expires"

  AUTHORIZATION = "authorization"
  CONTENT_MD5 = "content-md5"
  CONTENT_TYPE = "content-type"
  DATE = "date"
  RANGE = 'range'

  REQUEST_TIME_LIMIT = 900000  # 15min

  CACHE_CONTROL = "cache-control"
  CONTENT_ENCODING = "content-encoding"
  CONTENT_LENGTH = "content-length"
  LAST_MODIFIED = "last-modified"
  LAST_CHECKED = "last-checked"
  UPLOAD_TIME = "upload-time"
  CONTENT_RANGE = "content-range"
  EXPIRATION_TIME = "expiration-time"
  OBJECT_OWNER_ID = "object-owner-id"
  FILE_MODE = "file-mode"
  MULITPART_UPLOAD_MODE = XIAOMI_HEADER_PREFIX + "multipart-upload-mode"
  PREVIOUS_VERSION_ID = "pre-version-id"
  SERVER_SIDE_ENCRYPTION = XIAOMI_HEADER_PREFIX + "server-side-encryption"
  STORAGE_CLASS = XIAOMI_HEADER_PREFIX + "storage-class"
  ONGOING_RESTORE = XIAOMI_HEADER_PREFIX + "ongoing-restore"
  RESTORE_EXPIRY_DATE = XIAOMI_HEADER_PREFIX + "restore-expiry";

  DEFAULT_FDS_SERVICE_BASE_URI = "http://files.fds.api.xiaomi.com/"
  DEFAULT_CDN_SERVICE_URI = "http://cdn.fds.api.xiaomi.com/"



import json
import hashlib
from sys import version_info

from datetime import datetime

IS_PY3 = version_info[0] >= 3

if IS_PY3:
    import email.utils as rfc822
else:
    import rfc822


def rfc822_timestamp(time_string):
    return datetime.fromtimestamp(rfc822.mktime_tz(rfc822.parsedate_tz(time_string)))


def uri_to_bucket_and_object(uri):
    if not uri.startswith("fds://"):
        return None, None

    bucket_object_pair = uri[6:].split('/', 1)
    bucket = bucket_object_pair[0]
    object = bucket_object_pair[1]
    return bucket, object


def to_json_object(src):
    if isinstance(src, bytes):
        src = src.decode(encoding='utf-8')
    if src:
        return json.loads(src)
    return None

def file_md5(filename, blocksize=2**20):
    m = hashlib.md5()
    with open(filename, "rb") as f:
        while True:
            buf = f.read(blocksize)
            if not buf:
                break
            m.update(buf)
    return m.hexdigest()



from ..galaxy_fds_client_exception import GalaxyFDSClientException


class AccessControlPolicy(dict):
  '''
  The access control policy class.
  '''

  def __init__(self, json):
    '''
    Construct access control policy object from json.
    '''
    if json is not None:
      if 'owner' in json.keys():
        self.owner = json['owner']
      else:
        self.owner = None
      if 'accessControlList' in json.keys():
        self.access_control_list = json['accessControlList']
      else:
        self.access_control_list = None
    else:
      raise GalaxyFDSClientException("Json data cannot be None")

  @property
  def owner(self):
    return self.owner

  @owner.setter
  def owner(self, owner):
    self['owner'] = owner

  @property
  def access_control_list(self):
    return self.access_control_list

  @access_control_list.setter
  def access_control_list(self, access_control_list):
    self['accessControlList'] = access_control_list



from ..galaxy_fds_client_exception import GalaxyFDSClientException


class ListDomainMappingsResult(object):
  '''
  The List Domain Mappings Result class:
  '''

  def __init__(self, json):
    if json is not None:
      if 'domainMappings' in json.keys():
        self.domain_mappings = json['domainMappings']
      else:
        self.domain_mappings = None
    else:
      raise GalaxyFDSClientException("Json data cannot be None")




from .permission import Owner


class FDSBucket(object):
  """
  The FDS bucket class.
  """

  def __init__(self, bucket_name, owner=None):
    self.bucket_name = bucket_name
    self.owner = owner
    self.create_date = None

  def __str__(self):
    return self.bucket_name

  @property
  def owner(self):
    return self._owner

  @owner.setter
  def owner(self, json):
    if isinstance(json, dict):
      self._owner = Owner().from_json(json)
    else:
      self._owner = json



from .fds_client_configuration import FDSClientConfiguration
from .galaxy_fds_client import GalaxyFDSClient
from .galaxy_fds_client_exception import GalaxyFDSClientException
from .utils import rfc822_timestamp



class UploadPartResult(dict):
  """
  The Upload Part Result class.
  """

  def __init__(self, json):
    self.partNumber = json['partNumber']
    self.etag = json['etag']
    self.partSize = json['partSize']

  @property
  def partNumber(self):
    return self.partNumber

  @partNumber.setter
  def partNumber(self, part_number):
    self['partNumber'] = part_number

  @property
  def etag(self):
    return self.etag

  @etag.setter
  def etag(self, etag):
    self['etag'] = etag

  @property
  def partSize(self):
    return self.partSize

  @partSize.setter
  def partSize(self, part_size):
    self['partSize'] = part_size



from __future__ import print_function

import json


class FDSCORSConfig(dict):
  '''
  cors config like this:

  {
   "rules": [
     {
       "id": "0",
       "allowOrigin":"*.example.com"
     },
     {
       "id": "1",
       "allowOrigin":"*"
     }
   ]
  }
  '''

  def __init__(self, json={}):
    dict.__init__(self, json)
    self._rules = []
    for rule in self.get('rules', []):
      self._rules.append(FDSCORSRule(rule))
    self['rules'] = self._rules

  @property
  def rules(self):
    return self._rules

  def get_rule_by_id(self, id):
    for rule in self.rules:
      if rule.id == id:
        return rule
    return None

class FDSCORSRule(dict):
  def __init__(self, json = {}):
    dict.__init__(self, json)

  @property
  def id(self):
    return self.get('id', None)

  @id.setter
  def id(self, id):
    self['id'] = id

  @property
  def allowOrigin(self):
    return self.get('allowOrigin', None)

  @allowOrigin.setter
  def allowOrigin(self, allowOrigin):
    self['allowOrigin'] = allowOrigin



if __name__ == '__main__':
  cors_config = FDSCORSConfig()
  rule1 = FDSCORSRule()
  rule1.allowOrigin = '*.example.com'

  print(json.dumps(rule1, sort_keys=True))

  cors_config.rules.append(rule1)

  print(json.dumps(cors_config, sort_keys=True))

  cors_config.rules.append(rule1)
  print(json.dumps(cors_config, sort_keys=True))

  jsonstr = '''  {
   "rules": [
       {
         "id": "0",
         "allowOrigin":"*.example.com"
       },
       {
         "id": "1",
         "allowOrigin":"*"
       }
     ]
 }
  '''
  print(jsonstr)
  print(json.dumps(json.loads(jsonstr), sort_keys=True))
  print(json.dumps(FDSCORSConfig(json.loads(jsonstr)), sort_keys=True))



class Error(Exception):
  pass


class GalaxyFDSClientException(Error):
  def __init__(self, message):
    self.message = message

  def __str__(self):
    if self.message:
      return super.__str__(self) + ": " + str(self.message)
    return super.__str__(self)



class TimestampAntiStealingLinkConfig(dict):
  def __init__(self, json):
    if json is not None:
      if 'enabled' in json.keys():
        self.enabled = json['enabled']
      else:
        self.enabled = None
      if 'primaryKey' in json.keys():
        self.primary_key = json['primaryKey']
      else:
        self.primary_key = None
      if 'secondaryKey' in json.keys():
        self.secondary_key = json['secondaryKey']
      else:
        self.secondary_key = None
    else:
      raise GalaxyFDSClientException("Json data cannot be None")

  @property
  def enabled(self):
    return self['enabled']

  @enabled.setter
  def enabled(self, enabled):
    self['enabled'] = enabled

  @property
  def primary_key(self):
    return self['primaryKey']

  @primary_key.setter
  def primary_key(self, primary_key):
    self['primaryKey'] = primary_key

  @property
  def secondary_key(self):
    return self['secondaryKey']

  @secondary_key.setter
  def secondary_key(self, secondary_key):
    self['secondaryKey'] = secondary_key



import unittest

import utils


class UtilsTest(unittest.TestCase):

  def test_rui_to_bucket_and_object(self):
    uri = "http://bucket1/object1"
    bucket, object = utils.uri_to_bucket_and_object(uri)
    self.assertEquals(bucket, None)
    self.assertEquals(object, None)

    uri = "fds://bucket1/object1"
    bucket, object = utils.uri_to_bucket_and_object(uri)
    self.assertEquals(bucket, "bucket1")
    self.assertEquals(object, "object1")

    uri = "fds://bucket1/folder1/folder2/object1"
    bucket, object = utils.uri_to_bucket_and_object(uri)
    self.assertEquals(bucket, "bucket1")
    self.assertEquals(object, "folder1/folder2/object1")


if __name__ == "__main__":
  unittest.main()



import base64
import hmac
from email.utils import formatdate
from hashlib import sha1
from sys import version_info

from requests.auth import AuthBase

IS_PY3 = version_info[0] >= 3
if IS_PY3:
  from urllib.parse import unquote, urlparse
else:
  from urllib import unquote
  from urlparse import urlparse

from fds.auth.common import Common
from fds.model.subresource import SubResource


class Signer(AuthBase):
  '''
  The signer class used to sign the request.
  '''

  def __init__(self, app_key, app_secret, service_url=None):
    if service_url:
      self._service_base_url = service_url
    self._app_key = str(app_key)
    self._app_secret = str(app_secret)

  def __call__(self, request):
    request.headers[Common.DATE] = formatdate(timeval=None, localtime=False,
                                              usegmt=True)
    signature = self._sign_to_base64(request.method, request.headers,
                                     unquote(request.url), self._app_secret)
    request.headers[Common.AUTHORIZATION] = 'Galaxy-V2 %s:%s' % (
      self._app_key, signature)
    return request

  def _sign(self, method, headers, url, app_secret):
    '''
    Sign the specified http request.
    :param method:     The request method to sign
    :param headers:    The request headers to sign
    :param url:        The request uri to sign
    :param app_secret: The secret used to sign the request
    :return: The signed result, aka the signature
    '''
    string_to_sign = self._construct_string_to_sign(method, headers, url)
    if IS_PY3:
      if isinstance(app_secret, str):
        app_secret = app_secret.encode(encoding='utf-8')
      if isinstance(string_to_sign, str):
        string_to_sign = string_to_sign.encode(encoding='utf-8')
    digest = hmac.new(app_secret, string_to_sign, digestmod=sha1)
    return digest.digest()

  def _sign_to_base64(self, method, headers, url, app_secret):
    '''
    Sign the specified request to base64 encoded result.
    :param method:     The request method to sign
    :param headers:    The request headers to sign
    :param url:        The request uri to sign
    :param app_secret: The secret used to sign the request
    :return: The signed result, aka the signature
    '''
    signature = self._sign(method, headers, url, app_secret)
    if IS_PY3:
      return base64.encodebytes(signature).decode(encoding='utf-8').strip()
    else:
      return base64.encodestring(signature).strip()

  def _construct_string_to_sign(self, http_method, http_headers, uri):
    '''
    Construct the string used to sign the request.
    '''
    result = str()
    result += '%s\n' % http_method
    result += '%s\n' % self._get_header_value(http_headers,
                                              Common.CONTENT_MD5)
    result += '%s\n' % self._get_header_value(http_headers,
                                              Common.CONTENT_TYPE)
    expires = self._get_expires(uri)

    if expires > 0:
      result += '%s\n' % expires
    else:
      xiaomi_date = self._get_header_value(http_headers,
                                           Common.XIAOMI_HEADER_DATE)
      date = str()
      if len(xiaomi_date) == 0:
        date = self._get_header_value(http_headers, Common.DATE)
      result += '%s\n' % date

    result += '%s' % self._canonicalize_xiaomi_headers(http_headers)
    result += '%s' % self._canonicalize_resource(uri)
    return result

  def _get_header_value(self, http_headers, name):
    if http_headers is not None and name in http_headers:
      value = http_headers[name]
      if type(value) is list:
        return http_headers[name][0]
      else:
        return value
    return ""

  def _canonicalize_xiaomi_headers(self, http_headers):
    if http_headers is None or len(http_headers) == 0:
      return ''

    canonicalized_headers = dict()
    for key in http_headers:
      lower_key = key.lower()
      try:
        lower_key = lower_key.decode('utf-8')
      except:
        pass

      if http_headers[key] and lower_key.startswith(Common.XIAOMI_HEADER_PREFIX):
        if type(http_headers[key]) != str:
          canonicalized_headers[lower_key] = str()
          i = 0
          for k in http_headers[key]:
            canonicalized_headers[lower_key] += '%s' % (k.strip())
            i += 1
            if i < len(http_headers[key]):
              canonicalized_headers[lower_key] += ','
        else:
          canonicalized_headers[lower_key] = http_headers[key].strip()

    result = ""
    for key in sorted(canonicalized_headers.keys()):
      values = canonicalized_headers[key]
      result += '%s:%s\n' % (key, values)
    return result

  def _canonicalize_resource(self, uri):
    result = ""
    parsed_url = urlparse(uri)
    result += '%s' % parsed_url.path
    query_args = parsed_url.query.split('&')

    i = 0
    for query in sorted(query_args):
      key = query.split('=')
      if key[0] in SubResource.get_all_subresource():
        if i == 0:
          result += '?'
        else:
          result += '&'
        if len(key) == 1:
          result += '%s' % key[0]
        else:
          result += '%s=%s' % (key[0], key[1])
        i += 1
    return result

  def _get_expires(self, uri):
    parsed_url = urlparse(uri)
    query_args = sorted(parsed_url.query.split('&'))
    for query in query_args:
      key = query.split('=')[0]
      if key == Common.EXPIRES:
        return int(query.split('=')[1])
    return 0



from __future__ import print_function

import concurrent.futures
import errno
import json
import logging
import os
import mimetypes
import sys

try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path
from pprint import pprint
from sys import version_info

import click
from datetime import datetime


from fds import FDSClientConfiguration, GalaxyFDSClient, GalaxyFDSClientException
from fds.model.fds_object_metadata import FDSObjectMetadata
from fds.utils import rfc822_timestamp, file_md5
from fds.auth.common import Common


IS_PY3 = version_info[0] >= 3

if not IS_PY3:
    input = raw_input

log_format = "%(asctime)-15s [%(filename)s:%(lineno)d] %(message)s"
logging.basicConfig(format=log_format)
logger = logging.getLogger("fds.cli")
logger.setLevel(logging.INFO)

fds_prefix = r"fds://"

fds_client = None
fds_ak = None
fds_sk = None
fds_enable_https = None
fds_enable_cdn_for_download = None
fds_timeout = None
fds_part_size = None
fds_endpoint = None
fds_config = None


def mkdirs(path):
    try:
        if not os.path.isdir(path):
            os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(
            os.path.join(os.path.expanduser("~"), ".config", "xiaomi")
        ):
            pass


class WrongURLFormationException(Exception):
    pass


class WrongEnvironmentException(Exception):
    pass


class LocalConfig(object):
    def __init__(self):
        self.__config_path = os.path.join(
            os.path.expanduser("~"), ".config", "xiaomi", "config"
        )
        mkdirs(os.path.join(os.path.expanduser("~"), ".config", "xiaomi"))

        self.__data = None

        if not os.path.exists(self.__config_path):
            with open(self.__config_path, "w+") as f:
                f.writelines("{}")

        with open(self.__config_path, "r+") as f:
            self.__data = json.load(f)

    @property
    def ak(self):
        return self.__data.get("xiaomi_access_key_id")

    @ak.setter
    def ak(self, value):
        if value is not None and value.strip() != "":
            self.__data["xiaomi_access_key_id"] = value
            self.__dump()

    @property
    def sk(self):
        return self.__data.get("xiaomi_secret_access_key")

    @sk.setter
    def sk(self, value):
        if value is not None and value.strip() != "":
            self.__data["xiaomi_secret_access_key"] = value
            self.__dump()

    @property
    def endpoint(self):
        return self.__data.get("xiaomi_fds_endpoint")

    @endpoint.setter
    def endpoint(self, value):
        if value is not None and value.strip() != "":
            self.__data["xiaomi_fds_endpoint"] = value
            self.__dump()

    def __dump(self):
        with open(self.__config_path, "w") as outfile:
            json.dump(
                self.__data, outfile, sort_keys=True, indent=4, separators=(",", ": ")
            )


class CLIPrinter(object):
    def __init__(self):
        pass

    @staticmethod
    def done(message):
        print("[done] ", message)

    @staticmethod
    def warn(message):
        print("[warn] ", message)

    @staticmethod
    def fail(message):
        print("[fail] ", message)

    @staticmethod
    def print_bucket(bucket):
        print(bucket.bucket_name)

    @staticmethod
    def wrong_format():
        CLIPrinter.fail("wrong fds url format or local not exists")

    @staticmethod
    def print_object(object_name, metadata, human):
        content_length = int(metadata.metadata["x-xiaomi-meta-content-length"])
        if human == "k":
            content_length = content_length / 8.0 / 1024
        elif human == "m":
            content_length = content_length / 8.0 / 1024 / 1024
        elif human == "g":
            content_length = content_length / 8.0 / 1024 / 1024 / 1024
        print(
            "{}\t{:.2f}{}\t{}".format(
                metadata.metadata["last-modified"],
                content_length,
                human,
                str(object_name),
            )
        )

    @staticmethod
    def print_lifecycle(lifecycle):
        for action in lifecycle["rules"]:
            pprint(action)
            print("------------------------------------------------")


class FDSURL(object):
    def __init__(self, url):
        if not self.is_fds_url(url):
            raise WrongURLFormationException("Wrong fds url formation")

        self.__url = url
        self.__none_fds_prefix_name = url[len(fds_prefix) :]

    def is_bucket_url(self):
        """
    fds://bucket_name
    fds://bucket_name/
    """
        if (
            self.__none_fds_prefix_name == self.bucket_name()
            or self.__none_fds_prefix_name == self.bucket_name() + "/"
        ):
            return True
        return False

    def bucket_name(self):
        return self.__none_fds_prefix_name.split("/")[0]

    def object_name(self):
        if self.is_dir():
            return None
        if self.is_bucket_url():
            return None
        return self.__none_fds_prefix_name[len(self.bucket_name()) + 1 :]

    def file_name(self):
        if not self.is_object_url():
            return None

        return self.__none_fds_prefix_name.split("/")[-1]

    def is_bucket_dir(self):
        """
    fds://bucket_name/
    """
        return self.is_bucket_url() and self.is_dir()

    def is_object_dir(self):
        """
    fds://bucket_name/object_name/
    """
        return not self.is_bucket_url() and self.is_dir()

    def is_dir(self):
        """
    fds://bucket_name/
    fds://bucket_name/object_name/
    """
        return self.__url.endswith("/")

    def object_dir(self):
        if self.is_bucket_url():
            return None
        if not self.is_dir():
            return None
        return self.__none_fds_prefix_name[len(self.bucket_name()) + 1 :]

    def is_object_url(self):
        return not self.is_bucket_url() and not self.is_object_dir()

    @staticmethod
    def is_fds_url(url):
        if url is not None and len(url) > len(fds_prefix):
            return url.startswith(fds_prefix)
        return False

    @property
    def url(self):
        return self.__url


class FDSCliGroup(click.Group):
    def __call__(self, *args, **kwargs):
        try:
            return self.main(*args, **kwargs)
        except Exception as e:
            click.echo(str(e))


@click.group(cls=FDSCliGroup)
@click.option("--ak", default=None, help="Access Key ID")
@click.option("--sk", default=None, help="Access Key Secret")
@click.option("--endpoint", default=None, help="FDS Endpoint")
@click.option("--cdn_download", is_flag=True, help="Whether to download using cdn")
@click.option("--https", is_flag=True, help="Whether to download using https")
@click.option("--timeout", default=30, help="Client Timeout")
@click.option(
    "--part_size", default=50 * 1024 * 1024, help="Part size when multipart uploading"
)
@click.pass_context
def cli(ctx, ak, sk, endpoint, https, cdn_download, timeout, part_size):
    ctx.ensure_object(dict)
    global fds_client
    global fds_ak
    global fds_sk
    global fds_endpoint
    global fds_config

    if ctx.invoked_subcommand != "config":
        local_config = LocalConfig()
        env_ak = os.environ.get("FDS_AK")
        env_sk = os.environ.get("FDS_SK")
        env_endpoint = os.environ.get("FDS_ENDPOINT")

        ak = ak or env_ak or local_config.ak
        if not ak:
            raise WrongEnvironmentException(
                "Please setup ak, the loading order is: command line > environment > config file"
            )

        sk = sk or env_sk or local_config.sk
        if not sk:
            raise WrongEnvironmentException(
                "Please setup sk, the loading order is: command line > environment > config file"
            )

        endpoint = endpoint or env_endpoint or local_config.endpoint
        if not endpoint:
            raise WrongEnvironmentException(
                "Please setup endpoint, the loading order is: command line > environment > config file"
            )

        config = FDSClientConfiguration(
            region_name="awsde0",
            enable_https=https,
            enable_cdn_for_download=cdn_download,
            enable_cdn_for_upload=False,
            endpoint=endpoint,
            part_size=part_size,
            timeout=timeout,
        )
        fds_config = config
        fds_ak = ak
        fds_sk = sk
        fds_endpoint = endpoint
        fds_client = GalaxyFDSClient(access_key=ak, access_secret=sk, config=config)


@cli.command()
def config():
    """config ak, sk, endpoint and so on"""
    local_config = LocalConfig()
    default_ak = local_config.ak
    default_sk = local_config.sk
    default_endpoint = local_config.endpoint
    ak = input("enter access key id[default: %s]: " % default_ak)
    if ak == "":
        ak = default_ak
    sk = input("enter secret access key[default: %s]: " % default_sk)
    if sk == "":
        sk = default_sk
    endpoint = input("enter endpoint[default: %s]: " % default_endpoint)
    if endpoint == "":
        endpoint = default_endpoint

    local_config.ak = ak
    local_config.sk = sk
    local_config.endpoint = endpoint


@cli.command()
@click.argument("fds_url")
def mb(fds_url):
    """
  create(make) a bucket
  """
    url = FDSURL(fds_url)
    bucket_name = url.bucket_name()
    fds_client.create_bucket(bucket_name)
    click.echo("create bucket [%s]" % bucket_name)


@cli.command()
@click.argument("fds_url")
@click.option(
    "-f", "--force", help="Delete bucket although it is nonempty", is_flag=True
)
def rb(fds_url, force):
    """
  delete(remove) a bucket
  """
    url = FDSURL(fds_url)
    bucket_name = url.bucket_name()
    if force:
        result = fds_client.list_objects(bucket_name, "", "")
        while True:
            names = []
            for object_summary in result.objects:
                fds_client.delete_object(bucket_name, object_summary.object_name)
                click.echo(
                    "Deleted {}/{}".format(
                        url.bucket_name(), object_summary.object_name
                    )
                )
            if result.is_truncated:
                result = fds_client.list_next_batch_of_objects(result)
            else:
                break

    fds_client.delete_bucket(bucket_name)
    CLIPrinter.done("removed bucket [%s]" % bucket_name)


@cli.command()
@click.argument("fds_url", required=False)
@click.option("-r", "--recursive", default=False, help="Recursively list", is_flag=True)
@click.option("-h", "--human", default="k", help="human should be in 'k|m|g'")
def ls(fds_url, recursive, human):
    """
  list all buckets or objects in a bucket
  """
    if human != "k" and human != "m" and human != "g":
        raise WrongURLFormationException("human should be in 'k|m|g'")

    # bucket_url is None means listing all bucket name
    if fds_url is None:
        buckets = None
        buckets = fds_client.list_authorized_buckets()
        for bucket in buckets:
            click.echo(bucket)
    else:
        delimiter = "/"
        if recursive:
            delimiter = ""

        url = FDSURL(fds_url)
        prefix = url.object_name() or url.object_dir() or ""
        bucket_name = url.bucket_name()
        try:
            results = fds_client.list_all_objects(bucket_name, prefix, delimiter)
            for result in results:
                metadata = fds_client.get_object_metadata(
                    bucket_name, result.object_name
                )
                CLIPrinter.print_object(result.object_name, metadata, human)
        except GalaxyFDSClientException as e:
            click.echo(str(e))


@cli.command()
@click.argument("fds_url")
def rm(fds_url):
    """delete(remove) a object"""
    url = FDSURL(fds_url)

    if url.is_bucket_url():
        raise WrongURLFormationException(
            "The formation must be same as: fds://{bucketname}{objectname}"
        )
    bucket_name = url.bucket_name()
    object_name = url.object_name()

    fds_client.delete_object(bucket_name, object_name)
    click.echo("removed {}/{}".format(bucket_name, object_name))


@cli.command()
@click.argument("fds_url")
@click.option("-e", "--expires", default=1, help="Expiration time")
@click.option("-c", "--cdn", default=False, help="Use cdn url?", is_flag=True)
def presigned(fds_url, expires, cdn):
    """presigned command generates presigned url for download project"""
    url = FDSURL(fds_url)

    if url.is_bucket_url():
        raise WrongURLFormationException(
            "The formation must be same as: fds://{bucketname}/{objectname}"
        )

    bucket_name = url.bucket_name()
    object_name = url.object_name()

    expiration = int(
        1000
        * (
            float(expires) * 3600
            + float(
                (datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0, 0)).total_seconds()
            )
        )
    )

    if cdn:
        base_uri = fds_config.get_cdn_base_uri()
    else:
        base_uri = fds_config.get_base_uri()
    u = fds_client.generate_presigned_uri(
        base_uri, bucket_name, object_name, expiration
    )
    click.echo("generated presigned url: " + u)


@cli.command()
@click.argument("fds_url")
def show_ttl(fds_url):
    """ttl command shows the lifecycle information of a bucket or a object"""
    url = FDSURL(fds_url)

    bucket_name = url.bucket_name()
    ttl = fds_client.get_lifecycle_config(bucket_name)
    if url.is_bucket_url():
        CLIPrinter.print_lifecycle(ttl)
    elif url.is_object_url():
        if not fds_client.does_object_exists(bucket_name, url.object_name()):
            raise Exception("object {} is not exist".format(url.object_name()))
        if url.is_object_dir():
            prefix = url.object_dir()
        else:
            prefix = url.object_name()
        rules = [rule for rule in ttl["rules"] if rule["prefix"] in prefix]
        CLIPrinter.print_lifecycle({"rules": rules})
    else:
        CLIPrinter.wrong_format()


@cli.command()
@click.argument("fds_url")
@click.option("--public/--no-public", default=True, help="Public or not")
def set_public(fds_url, public):
    """set the resource of fds public or not"""
    url = FDSURL(fds_url)
    if public:
        fds_client.set_public(url.bucket_name(), url.object_name())
    else:
        if url.is_object_dir():
            return
        fds_client.set_private(url.bucket_name(), url.object_name())


@cli.command()
@click.argument("fds_url")
@click.option("--out/--no-out", default=True, help="Outside access")
def access(fds_url, out):
    """set the accessibility of resource"""
    url = FDSURL(fds_url)
    if url.is_bucket_url():
        fds_client.set_bucket_outside_access(url.bucket_name(), out)
    elif url.is_object_url():
        fds_client.set_object_outside_access(url.bucket_name(), url.object_name(), out)
    else:
        raise WrongURLFormationException("Wrong url formation")


@cli.command()
def info():
    """display the configurations"""
    print("Access Key ID: {}".format(fds_ak))
    print("Access Secret Key: {}".format(fds_sk))
    print("Endpoint: {}".format(fds_endpoint))


def _cp(src_url, dst_url):
    src_bucket_name = src_url.bucket_name()
    src_object_name = src_url.object_name()

    dst_bucket_name = dst_url.bucket_name()

    if dst_url.is_object_url():
        dst_object_name = dst_url.object_name()
    else:
        dst_object_name = src_object_name
    fds_client.copy_object(
        src_bucket_name, src_object_name, dst_bucket_name, dst_object_name
    )
    click.echo(
        "copy {}/{} to {}/{}".format(
            src_bucket_name, src_object_name, dst_bucket_name, dst_object_name
        )
    )


def _cp_batch(src_url, dst_url, recursive):
    src_bucket_name = src_url.bucket_name()
    dst_bucket_name = dst_url.bucket_name()

    prefix = ""
    if src_url.is_object_dir():
        prefix = src_url.object_dir()

    delimiter = "/"
    if recursive:
        delimiter = ""

    all_objects = fds_client.list_all_objects(
        bucket_name=src_bucket_name, prefix=prefix, delimiter=delimiter
    )
    for o in all_objects:
        o_name = o.object_name
        fds_client.copy_object(src_bucket_name, o_name, dst_bucket_name, o_name)
        click.echo(
            "copy {}/{} to {}/{}".format(
                src_bucket_name, o_name, dst_bucket_name, o_name
            )
        )


def _download(src_url, dst):
    src_bucket_name = src_url.bucket_name()
    src_object_name = src_url.object_name()

    if os.path.isdir(dst):
        if dst == "." or dst == "..":
            dst_name = src_url.file_name()
        elif dst.endswith("/"):
            dst_name = dst + src_url.file_name()
        else:
            dst_name = dst + "/" + src_object_name.split("/")[-1]
    else:
        dst_name = dst
    mtime = None
    if os.path.isfile(dst_name):
        local_md5 = file_md5(dst_name)
        remote_md5 = fds_client.get_object_metadata(
            src_bucket_name, src_object_name
        ).metadata.get(Common.CONTENT_MD5)
        if remote_md5 is not None and local_md5 == remote_md5:
            CLIPrinter.done(
                "download %s/%s to local(skip because of same md5)"
                % (src_bucket_name, src_object_name)
            )
            return

        mtime = os.path.getmtime(dst_name)

    try:
        fds_object = fds_client.get_object(
            bucket_name=src_bucket_name, object_name=src_object_name, stream=True
        )
    except GalaxyFDSClientException as e:
        CLIPrinter.fail(e.message)
        return
    lm = fds_object.metadata.metadata["last-modified"]
    remote_modified = rfc822_timestamp(lm)

    # if last-modified of local file is not less last-modified of remote file, skip
    if mtime is not None and datetime.fromtimestamp(mtime) >= remote_modified:
        CLIPrinter.done(
            "download %s/%s to local(skip because of updated)"
            % (src_bucket_name, src_object_name)
        )
        return

    length_left = IS_PY3 and sys.maxsize or sys.maxint
    try:
        with open(dst_name, "wb") as f:
            for chunk in fds_object.stream:
                length = min(length_left, len(chunk))
                f.write(chunk[0:length])
                length_left -= length
                if length_left <= 0:
                    break
    except Exception as exception:
        print(exception)
    finally:
        fds_object.stream.close()
    CLIPrinter.done("download %s/%s to local" % (src_bucket_name, src_object_name))


def _upload(fpath, object_name, dst_url, autodetect_mimetype):
    if not fpath.exists():
        CLIPrinter.warn("{} is a bad file".format(str(fpath)))
        return

    try:
        dst_bucket_name = dst_url.bucket_name()
        dst_object_name = object_name
        if fds_client.does_object_exists(dst_bucket_name, dst_object_name):
            # check md5 firstly
            metadata = fds_client.get_object_metadata(dst_bucket_name, dst_object_name)
            if metadata.metadata.get(Common.CONTENT_MD5) is not None:
                local_md5 = file_md5(str(fpath.resolve()))
                if local_md5 == metadata.metadata.get(Common.CONTENT_MD5):
                    CLIPrinter.done(
                        "upload object %s/%s(skip because of same md5)"
                        % (dst_bucket_name, dst_object_name)
                    )
                    return

            # check last-modified
            mtime = None
            if fpath.is_file():
                mtime = os.path.getmtime(str(fpath.resolve()))

            lm = metadata.metadata[Common.LAST_MODIFIED]
            remote_modified = rfc822_timestamp(lm)

            # if last-modified of local file is not less last-modified of remote file, skip
            if mtime is not None and datetime.fromtimestamp(mtime) <= remote_modified:
                CLIPrinter.done(
                    "upload object %s/%s(skip because of updated)"
                    % (dst_bucket_name, dst_object_name)
                )
                return
    except Exception as e:
        CLIPrinter.fail(e.message)
        return
    mimetype = None
    if autodetect_mimetype:
        mimetype = mimetypes.guess_type(str(fpath.resolve()))[0]
    metadata = FDSObjectMetadata()
    if mimetype is not None:
        metadata.add_header(Common.CONTENT_TYPE, mimetype)

    with open(str(fpath.resolve()), "rb") as f:
        try:
            fds_client.put_object(
                dst_bucket_name, dst_object_name, f, metadata=metadata
            )
            CLIPrinter.done("upload object %s/%s" % (dst_bucket_name, dst_object_name))
        except GalaxyFDSClientException as e:
            CLIPrinter.fail("upload object %s/%s, %s" % (dst_bucket_name, dst_object_name, e.message))


def _upload_batch(s_path, dst_url, concurrency, autodetect_mimetype):
    first_object_name = dst_url.object_dir() or ""
    with concurrent.futures.ThreadPoolExecutor(max_workers=concurrency) as executor:
        for fpath in s_path.glob("**/*"):
            extra_object_name = fpath.relative_to(s_path)
            object_name = str(first_object_name / extra_object_name)
            object_name = "/".join(object_name.split("\\"))
            if fpath.is_file():
                executor.submit(
                    _upload, fpath, object_name, dst_url, autodetect_mimetype
                )


@cli.command()
@click.argument("src")
@click.argument("dst")
@click.option(
    "-a", "--autodetect_mimetype", is_flag=True, help="Whether to autodetect mimetype"
)
def cp(src, dst, autodetect_mimetype):
    """cp command do lots of things.

    \b
    1. rename a fds file from a to b
    2. upload a local file to fds
    3. download a fds file to local
    4. copy a fds file from bucket1 to bucket2
    5. can not handle directory
    """

    if FDSURL.is_fds_url(src) and FDSURL.is_fds_url(dst):
        src_url = FDSURL(src)
        dst_url = FDSURL(dst)

        if not src_url.is_object_url():
            raise WrongURLFormationException(
                "cp command can only handle an existing file"
            )
        _cp(src_url, dst_url)

    elif FDSURL.is_fds_url(src) and not FDSURL.is_fds_url(dst):
        src_url = FDSURL(src)

        if not src_url.is_object_url():
            raise WrongURLFormationException(
                "cp command can only handle an existing file"
            )
        _download(src_url, dst)

    elif not FDSURL.is_fds_url(src) and FDSURL.is_fds_url(dst):
        dst_url = FDSURL(dst)

        if not os.path.isfile(src):
            raise WrongURLFormationException(
                "cp command can only handle an existing file"
            )
        fpath = Path(src)
        if dst_url.is_object_url():
            object_name = dst_url.object_name()
        else:
            object_name = (dst_url.object_dir() or "") + fpath.name
        _upload(
            fpath.resolve(),
            object_name,
            dst_url,
            autodetect_mimetype=autodetect_mimetype,
        )
    else:
        CLIPrinter.fail("don't support copy file from local to local")


@cli.command()
@click.argument("src")
@click.argument("dst")
@click.option(
    "-a", "--autodetect_mimetype", is_flag=True, help="Whether to autodetect mimetype"
)
@click.option("--concurrency", default=1, help="counts of threads")
def sync(src, dst, autodetect_mimetype, concurrency):
    """sync command syncs between (local directory and fds) (fds and local directory)"""

    if FDSURL.is_fds_url(src) and not FDSURL.is_fds_url(dst):
        src_url = FDSURL(src)
        if src_url.is_object_url():
            raise WrongURLFormationException("sync can not handle file")

        src_bucket_name = src_url.bucket_name()
        prefix = src_url.object_dir() or ""
        try:
            all_objects = fds_client.list_all_objects(
                bucket_name=src_bucket_name, prefix=prefix, delimiter=""
            )
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=concurrency
            ) as executor:
                for o in all_objects:
                    o_name = o.object_name[len(prefix):]
                    o_name = o_name[1:] if o_name.startswith("/") else o_name
                    url = FDSURL(fds_prefix + src_bucket_name + "/" + o.object_name)
                    if "/" not in o_name:
                        executor.submit(_download, url, dst)
                    elif url.is_object_url():
                        o_file_name = o_name.split("/")[-1]
                        o_dir = str(Path(dst) / o_name.split(o_file_name)[0])
                        mkdirs(o_dir)
                        executor.submit(_download, url, o_dir)
        except GalaxyFDSClientException as e:
            CLIPrinter.fail(e.message)

    elif not FDSURL.is_fds_url(src) and FDSURL.is_fds_url(dst):
        dst_url = FDSURL(dst)
        s_path = Path(src)
        _upload_batch(s_path.resolve(), dst_url, concurrency, autodetect_mimetype)

    elif FDSURL.is_fds_url(src) and FDSURL.is_fds_url(dst):
        raise WrongURLFormationException(
            "sync command syncs between (local directory and fds) (fds and local directory)"
        )


def main():
    cli(obj={})


if __name__ == "__main__":
    main()




# wrap the requests library

import requests
from requests.sessions import HTTPAdapter


class FDSRequest:
  def __init__(self, timeout, max_retries):
    self._max_retries = max_retries
    self._timeout = timeout

  def request(self, method, url, kwargs):
    '''
    Constructs a specified session and sends quest.
    Returns :class:`Response <Response>` object.
    '''

    kwargs.setdefault('timeout', self._timeout)
    session = requests.Session()
    session.mount("http://", HTTPAdapter(max_retries=self._max_retries))
    session.mount("https://", HTTPAdapter(max_retries=self._max_retries))
    response = session.request(method=method, url=url, **kwargs)
    # By explicitly closing the session, we avoid leaving sockets open which
    # can trigger a ResourceWarning in some cases, and look like a memory leak
    # in others.
    session.close()
    return response

  def get(self, url, **kwargs):
    kwargs.setdefault('allow_redirects', True)
    return self.request('get', url, kwargs)

  def options(self, url, **kwargs):
    kwargs.setdefault('allow_redirects', True)
    return self.request('options', url, kwargs)

  def head(self, url, **kwargs):
    kwargs.setdefault('allow_redirects', False)
    return self.request('head', url, kwargs)

  def post(self, url, data=None, json=None, **kwargs):
    kwargs['data'] = data
    kwargs['json'] = json
    return self.request('post', url, kwargs)

  def put(self, url, data=None, **kwargs):
    kwargs['data'] = data
    return self.request('put', url, kwargs)

  def patch(self, url, data=None, **kwargs):
    kwargs['data'] = data
    return self.request('patch', url, kwargs)

  def delete(self, url, **kwargs):
    return self.request('delete', url, kwargs)



from .fds_object_summary import FDSObjectSummary
from .permission import Owner


class FDSObjectListing(dict):
  '''
  The FDS Object Listing class.
  '''

  def __init__(self, json):
    dict.__init__(self, json)
    self._objects = []
    for obj in self['objects']:
      summary = FDSObjectSummary()
      summary.bucket_name = self['name']
      summary.object_name = obj['name']
      summary.owner = Owner().from_json(obj['owner'])
      summary.size = obj['size']
      self._objects.append(summary)

  def __str__(self):
    return str(map(lambda x: x.object_name, self._objects))

  @property
  def prefix(self):
    return self['prefix']

  @prefix.setter
  def prefix(self, prefix):
    self['prefix'] = prefix

  @property
  def delimiter(self):
    return self['delimiter'];

  @delimiter.setter
  def delimiter(self, delimiter):
    self['delimiter'] = delimiter

  @property
  def marker(self):
    return self['marker']

  @marker.setter
  def marker(self, marker):
    self['marker'] = marker

  @property
  def next_marker(self):
    return self['nextMarker']

  @next_marker.setter
  def next_marker(self, next_marker):
    self['nextMarker'] = next_marker

  @property
  def bucket_name(self):
    return self['name']

  @bucket_name.setter
  def bucket_name(self, bucket_name):
    self['name'] = bucket_name

  @property
  def delimiter(self):
    return self['delimiter']

  @delimiter.setter
  def delimiter(self, delimiter):
    self['delimiter'] = delimiter

  @property
  def max_keys(self):
    return self['maxKeys']

  @max_keys.setter
  def max_keys(self, max_keys):
    self['maxKeys'] = max_keys

  @property
  def objects(self):
    return self._objects

  @objects.setter
  def objects(self, objects):
    objs = []
    for x in objects:
      if isinstance(x, FDSObjectSummary):
        raise TypeError("Parameter should be a list of FDSObjectSummary")
      objs.append(x)
    self._objects = objs

  @property
  def common_prefixes(self):
    return self['commonPrefixes']

  @common_prefixes.setter
  def common_prefixes(self, common_prefixes):
    self['commonPrefixes'] = common_prefixes

  @property
  def is_truncated(self):
    return self['truncated']

  @is_truncated.setter
  def is_truncated(self, is_truncated):
    self['truncated'] = is_truncated



class UploadPartResultList(dict):
  """
  The Upload Part Result List class
  """

  def __init__(self, json):
    if json:
      self.uploadPartResultList = json['uploadPartResultList']

  @property
  def uploadPartResultList(self):
    return self.uploadPartResultList

  @uploadPartResultList.setter
  def uploadPartResultList(self, upload_part_result_list):
    self['uploadPartResultList'] = upload_part_result_list



import unittest

from fds.fdscli_cli import FDSURL


class TestFDSURL(unittest.TestCase):
  def setUp(self):
    self.a = FDSURL("fds://bucket_name")
    self.b = FDSURL("fds://bucket_name/")
    self.c = FDSURL("fds://bucket_name/object_prefix/object_name")
    self.d = FDSURL("fds://bucket_name/prefix/dir_name/")

  def test_is_fds_url(self):
    self.assertTrue(FDSURL.is_fds_url("fds://bucket_name"))
    self.assertFalse(FDSURL.is_fds_url("dd/fdsurl"))

  def test_is_absolute_bucket_url(self):
    self.assertTrue(self.a.is_bucket_url())
    self.assertTrue(self.b.is_bucket_url())
    self.assertFalse(self.c.is_bucket_url())
    self.assertFalse(self.d.is_bucket_url())

  def test_bucket_name(self):
    self.assertEqual(self.a.bucket_name(), "bucket_name")
    self.assertEqual(self.b.bucket_name(), "bucket_name")
    self.assertEqual(self.c.bucket_name(), "bucket_name")
    self.assertEqual(self.d.bucket_name(), "bucket_name")

  def test_object_name(self):
    self.assertIsNone(self.a.object_name())
    self.assertIsNone(self.b.object_name())

    self.assertEqual(self.c.object_name(), "object_prefix/object_name")
    self.assertIsNone(self.d.object_name())

  def test_is_dir(self):
    self.assertFalse(self.a.is_dir())
    self.assertTrue(self.b.is_dir())
    self.assertFalse(self.c.is_dir())
    self.assertTrue(self.d.is_dir())

  def test_is_bucket_dir(self):
    self.assertFalse(self.a.is_bucket_dir())
    self.assertTrue(self.b.is_bucket_dir())
    self.assertFalse(self.c.is_bucket_dir())
    self.assertFalse(self.d.is_bucket_dir())

  def test_is_object_dir(self):
    self.assertFalse(self.a.is_object_dir())
    self.assertFalse(self.b.is_object_dir())
    self.assertFalse(self.c.is_object_dir())
    self.assertTrue(self.d.is_object_dir())

  def test_object_dir(self):
    self.assertIsNone(self.a.object_dir())
    self.assertIsNone(self.b.object_dir())
    self.assertIsNone(self.c.object_dir())
    self.assertEqual(self.d.object_dir(), "prefix/dir_name/")

  def test_is_object_url(self):
    self.assertFalse(self.a.is_object_url())
    self.assertFalse(self.b.is_object_url())
    self.assertTrue(self.c.is_object_url())
    self.assertFalse(self.d.is_object_url())


if __name__ == "__main__":
  unittest.main()



class FDSClientConfiguration(object):
  URI_HTTP = 'http://'

  URI_HTTPS = 'https://'

  URI_CDN = 'cdn'

  URI_SUFFIX = 'fds.api.xiaomi.com'

  URI_NET_SUFFIX = 'fds.api.xiaomi.net'

  URI_CDN_SUFFIX = 'fds.api.mi-img.com'

  def __init__(self, region_name='cnbj0',
               enable_cdn_for_download=True,
               enable_cdn_for_upload=False,
               enable_https=True,
               threshold_size=100 * 1024 * 1024,
               part_size=50 * 1024 * 1024,
               timeout=30,
               max_retries=3,
               endpoint=None):
    """
    :param region_name:
    :param enable_cdn_for_download:
    :param enable_cdn_for_upload:
    :param enable_https:
    :param timeout: connection and read timeout (seconds)
    :param max_retries:
    """
    if region_name is None:
      self._region_name = 'cnbj0'
    else:
      self._region_name = region_name
    self._enable_cdn_for_download = enable_cdn_for_download
    self._enable_cdn_for_upload = enable_cdn_for_upload
    self._enable_https = enable_https
    self._enable_md5_calculate = False
    self._timeout = timeout
    self._max_retries = max_retries
    self._debug = False
    if endpoint is None:
      endpoint = self._region_name + '.' + self.URI_SUFFIX

    self.set_endpoint(endpoint)

    self._threshold_size = threshold_size
    self._part_size = part_size

  def get_threshold_size(self):
    return self._threshold_size

  def set_threshold_size(self, threshold_size):
    self._threshold_size = threshold_size

  def get_part_size(self):
    return self._part_size

  def set_part_size(self, part):
    self._part_size = part

  @property
  def debug(self):
    return self._debug

  @debug.setter
  def debug(self, debug):
    self._debug = debug

  @property
  def enable_md5_calculate(self):
    return self._enable_md5_calculate

  @enable_md5_calculate.setter
  def enable_md5_calculate(self, enable):
    self._enable_md5_calculate = enable

  @property
  def timeout(self):
    return self._timeout

  @timeout.setter
  def timeout(self, timeout):
    self._timeout = timeout

  @property
  def max_retries(self):
    return self._max_retries

  @max_retries.setter
  def max_retries(self, max_retries):
    self._max_retries = max_retries

  def set_endpoint(self, endpoint):
    assert endpoint.endswith(self.URI_SUFFIX) or endpoint.endswith(self.URI_NET_SUFFIX)

    self._endpoint = endpoint
    suffix = self._endpoint[-19:]
    i = self._endpoint.index(suffix)
    self._region_name = self._endpoint[0:i]
    self._cdn_endpoint = self.URI_CDN + '.' + self._region_name + '.' + self.URI_CDN_SUFFIX

  def get_download_base_uri(self):
    return self._build_base_uri(self._enable_cdn_for_download)

  def get_upload_base_uri(self):
    return self._build_base_uri(self._enable_cdn_for_upload)

  def get_base_uri(self):
    return self._build_base_uri(False)

  def get_cdn_base_uri(self):
    return self._build_base_uri(True)

  def _build_base_uri(self, enable_cdn):
    base_uri = str()
    if self._enable_https:
      base_uri += self.URI_HTTPS
    else:
      base_uri += self.URI_HTTP

    if enable_cdn:
      base_uri += self._cdn_endpoint
    else:
      base_uri += self._endpoint

    base_uri += '/'
    return base_uri



class InitMultipartUploadResult(object):
  """
  The Init Multipart Upload Result class:
  """

  def __init__(self, json):
    self.bucket_name = json['bucketName']
    self.object_name = json['objectName']
    self.upload_id = json['uploadId']



#!/usr/bin/env python
"""

Prerequest:
put following json into ~/.config/xiaomi/config
{
...
"xiaomi_access_key_id":"ACCESS_KEY",
"xiaomi_secret_access_key":"SECRET_KEY",
"xiaomi_fds_endpoint":"END_POINT"
, ...
}

Usage Examples:
\t[create bucket]\n\t\tfds -m put -b BUCKET_NAME
\t[list buckets]\n\t\tfds -l
\t[list objects under bucket]\n\t\tfds -l -b BUCKET_NAME
\t[list directory under bucket]\n\t\tfds -L DIR -b BUCKET_NAME
\t[create object under bucket]\n\t\tfds -m put -b BUCKET_NAME -o OBJECT_NAME -d FILE_PATH
\t[create object with pipline]\n\t\tcat file | fds -m put -b BUCKET_NAME -o OBJECT_NAME
\t[generate presigned url for upload, expires in 7 days]\n\t\tfds -p -m put -b BUCKET_NAME -o OBJECT_NAME --expiration '7*24' --metadata 'content-type:application/octet-stream'
\t\tYOUR_PRESIGNED_URL_GENRARTED
\t\t[use presigned url to upload object]\n\t\t\tcurl -v -X PUT -H 'Content-Type:application/octet-stream' 'YOUR_PRESIGNED_URL_GENRARTED' -d 'object content'
\t[generate presigned url for download, expires in 30 min]\n\t\tfds -p -m get -b BUCKET_NAME -o OBJECT_NAME --expiration '0.5'
\t\tYOUR_PRESIGNED_URL_GENRARTED
\t\t[use presigned url to download object]\n\t\t\twget -v 'YOUR_PRESIGNED_URL_GENRARTED'
\t[put bucket acl]\n\t\tfds -m put -b BUCKET_NAME --gratee AUTHENTICATED_USERS --permission READ
\t[delete object]\n\t\tfds -m delete -b BUCKET_NAME -o OBJECT_NAME [--disable_trash]
\t[delete objects]\n\t\tfds -m delete -b BUCKET_NAME --object_prefix=OBJECT_NAME_PREFIX [--disable_trash]
\t[delete empty bucket]\n\t\tfds -m delete -b BUCKET_NAME
\t[delete bucket with object]\n\t\tfds -m delete -b BUCKET_NAME --force
\t[download object]\n\t\tfds -m get -b BUCKET_NAME -o OBJECT_NAME -d OUTPUT_FILE_NAME
\t[download directory(object preifx) recursively]\n\t\tfds -m get -b BUCKET_NAME --P OBJECT_PREFIX -D OUTPUT_DIRECTORY -R
\t[download all objects under a bucket]\n\t\tfds -m get -b BUCKET_NAME -P / -D OUTPUT_DIRECTORY -R
\t[copy src_object from src_bucket to dst_bucket and rename it to dst_object]\n\t\tfds -m put -srcb SRC_BUCKET_NAME -srco SRC_OBJECT_NAME -dstb DST_BUCKET_NAME -dsto DST_OBJECT_NAME
"""
from __future__ import print_function

import json
import logging
import os
import sys
import traceback
from os.path import expanduser
from sys import version_info

import argcomplete
import argparse
from argcomplete.completers import ChoicesCompleter
from datetime import datetime
from time import sleep

from fds import FDSClientConfiguration, GalaxyFDSClient
from fds.galaxy_fds_client_exception import GalaxyFDSClientException
from fds.model.fds_lifecycle import FDSLifecycleConfig, FDSLifecycleRule
from fds.model.fds_cors import FDSCORSConfig, FDSCORSRule
from fds.model.fds_object_metadata import FDSObjectMetadata
from fds.model.permission import AccessControlList
from fds.model.permission import Grant
from fds.model.permission import GrantType
from fds.model.permission import Grantee
from fds.model.permission import Permission
from fds.model.permission import UserGroups
from fds.model.upload_part_result_list import UploadPartResultList

IS_PY3 = version_info[0] >= 3

if IS_PY3:
  pass
else:
  pass

logger = None
access_key = None
secret_key = None
region = ''
fds_config = None
enable_https = True
enable_cdn = False
end_point = None
fds_client = None
presigned_url = False
force_delete = False
disable_trash = False
object_prefix = None
gratee = None
permission = None
is_public = False
lifecycle = None
lifecycle_rule = None
cors=None
cors_rule=None
recursive = False
webp_quality = None
gif_extract_type = None
restore_archive = False
is_archive = False
list_trash = False

expiration_in_hour = '1.0'
multipart_upload_threshold_size = 50 * 1024 * 1024
multipart_upload_buffer_size = 10 * 1024 * 1024
max_upload_retry_time = 5


def print_config(name, value):
  global logger
  if not (logger is None):
    logger.debug('Config param [' + str(name) + '] ' + ' value [' + str(value) + ']')


def read_local_config():
  global logger
  config_dirs = [os.path.join(expanduser("~"), ".config", "xiaomi", "config")
    , os.path.join(expanduser("~"), ".config", "fds", "client.config")];
  for config_dir in config_dirs:
    if not os.path.exists(config_dir):
      if not (logger is None):
        logger.debug("local config not exist [" + str(config_dir) + "]")
    else:
      if not (logger is None):
        logger.debug("use config from [" + str(config_dir) + "]")
      with open(config_dir) as f:
        return json.load(fp=f)
  return {}


def parse_argument(args):
  global method, bucket_name, region, object_name, \
    enable_cdn, enable_https, list_dir, list_objects, \
    data_file, data_dir, start_mark, metadata, length, offset, \
    access_key, secret_key, end_point, presigned_url, expiration_in_hour, \
    force_delete, gratee, permission, disable_trash, object_prefix, lifecycle, lifecycle_rule, \
    recursive, dst_bucket_name, dst_object_name, src_bucket_name, src_object_name, webp_quality, gif_extract_type, cors, cors_rule,\
    restore_archive, is_archive, list_trash, is_public
  local_config = read_local_config()
  method = args.method
  print_config('method', method)
  if args.bucket:
    bucket_name = args.bucket
  else:
    bucket_name = local_config.get('bucket')
  print_config('bucket name', bucket_name)
  region = local_config.get('region')
  if args.region:
    region = args.region
    if args.end_point:
      end_point = args.end_point
  else:
    if args.end_point:
      end_point = args.end_point
    else:
      end_point = local_config.get('xiaomi_fds_endpoint')
      if end_point is None:
        end_point = local_config.get('end_point')

  print_config('region name', region)
  object_name = args.object
  print_config('object name', object_name)
  enable_cdn = args.CDN
  print_config('cdn enabled', enable_cdn)

  if type(args.https) == str:
    enable_https = args.https.lower() == "true"
  else:
    enable_https = local_config.get("xiaomi_fds_https", True)

  print_config('https enabled', enable_https)
  list_dir = args.list_dir
  print_config('list dir', list_dir)
  list_objects = args.list_objects
  print_config('list objects', list_objects)
  data_file = args.data_file
  print_config('data file', data_file)
  data_dir = args.data_dir
  print_config('data directory', data_dir)
  start_mark = args.start_mark
  print_config('start mark', start_mark)
  metadata = args.metadata
  print_config('meta data', metadata)
  length = args.length
  print_config('length', length)
  offset = args.offset
  print_config('offset', offset)
  print_config('end point', end_point)

  force_delete = args.force_delete
  print_config('force', force_delete)

  disable_trash = args.disable_trash
  print_config('disable_trash', disable_trash)

  object_prefix = args.object_prefix

  print_config('object_prefix', object_prefix)

  webp_quality = args.webp_quality
  print_config('webp_quality', webp_quality)

  gif_extract_type = args.gif_extract_type
  print_config('gif_extract_type', gif_extract_type)

  restore_archive = args.restore_archive
  print_config('restore archived object', restore_archive)

  is_archive = args.is_archive
  print_config('is_archive', is_archive)

  list_trash = args.list_trash
  print_config('list_trash', list_trash)

  if args.ak:
    access_key = args.ak
  else:
    if access_key is None and "XIAOMI_ACCESS_KEY_ID" in os.environ:
      access_key = os.environ["XIAOMI_ACCESS_KEY_ID"]
    if access_key is None:
      access_key = local_config.get('xiaomi_access_key_id')
    if access_key is None:
      access_key = local_config.get('ak')
  print_config('access key', access_key)
  if access_key is None:
    sys.stderr.write('Access key is not set, use command "fds -h" get help')
    sys.stderr.flush()
    exit(1)
  if args.sk:
    secret_key = args.sk
  else:
    if secret_key is None and "XIAOMI_SECRET_ACCESS_KEY" in os.environ:
      secret_key = os.environ["XIAOMI_SECRET_ACCESS_KEY"]
    if secret_key is None:
      secret_key = local_config.get('xiaomi_secret_access_key')
    if secret_key is None:
      secret_key = local_config.get('sk')
  if secret_key is None:
    sys.stderr.write('Secret key is not set, use command "fds -h" get help')
    sys.stderr.flush()
    exit(1)
  presigned_url = args.presigned_url
  print_config('presigned url', presigned_url)
  expiration_in_hour = args.expiration_in_hour
  print_config('expiration in hour', expiration_in_hour)
  gratee = args.gratee
  print_config('gratee', gratee)
  permission = args.permission
  print_config('permission', permission)
  is_public = args.is_public
  print_config('is_public', is_public)

  src_bucket_name = args.src_bucket_name
  print_config('src_bucket_name', src_bucket_name)
  src_object_name = args.src_object_name
  print_config('src_object_name', src_object_name)
  dst_bucket_name = args.dst_bucket_name
  print_config('dst_bucket_name', dst_bucket_name)
  dst_object_name = args.dst_object_name
  print_config('dst_object_name', dst_object_name)

  lifecycle = args.lifecycle
  print_config('lifecycle', lifecycle)
  lifecycle_rule = args.lifecycle_rule
  print_config('lifecycle_rule', lifecycle_rule)
  cors = args.cors
  print_config('cors', cors)
  cors_rule = args.cors_rule
  print_config('cors_rule', cors_rule)
  recursive = args.recursive
  print_config('recursive', recursive)


def get_buckets(fds_client):
  buckets = fds_client.list_buckets()
  bucket_names = {};
  for i in buckets:
    bucket_names[i.bucket_name] = ''
  authorized_buckets = fds_client.list_authorized_buckets()
  for i in authorized_buckets:
    if i.bucket_name not in bucket_names:
      buckets.append(i);
  return buckets


def list_buckets(fds_client, prefix, start_mark):
  buckets = get_buckets(fds_client)
  for i in buckets:
    if i.bucket_name.startswith(prefix) and (not start_mark or i.bucket_name >= start_mark):
      sys.stdout.write(i.bucket_name + '/')
      sys.stdout.write('\n')


def bucket_name_completer(prefix, parsed_args, **kwargs):
  parse_argument(args=parsed_args)

  if not (access_key is None) and not (secret_key is None) and not (region is None):
    argcomplete.warn(str(enable_https) + ' ' + str(enable_cdn) + ' ' + str(region))
    fds_config = FDSClientConfiguration(region_name=region,
                                        enable_https=enable_https,
                                        enable_cdn_for_download=enable_cdn,
                                        enable_cdn_for_upload=enable_cdn)
    fds_client = GalaxyFDSClient(access_key=access_key,
                                 access_secret=secret_key,
                                 config=fds_config)
    bucket_list = get_buckets(fds_client=fds_client)
    rtn = []
    for i in bucket_list:
      if i.startswith(prefix):
        rtn.append(i)
    return rtn
  return ['a', 'b', 'c']


def check_region(region):
  pass


def check_bucket_name(bucket_name):
  pass


def check_object_name(object_name):
  pass


def check_metadata(metadata):
  pass


def put_directory(data_dir, bucket_name, object_name_prefix, metadata):
  check_bucket_name(bucket_name)
  check_object_name(object_name_prefix)
  check_metadata(metadata)
  # The object name doesn't have starting '/'
  object_name_prefix = object_name_prefix.lstrip('/')
  for root, dirs, files in os.walk(data_dir):
    relative_dir = os.path.relpath(root, data_dir)
    for filename in files:
      data_file = os.path.join(root, filename)
      if filename.startswith('.'):
        logger.warn("object name can't start with '.', skipping " + data_file)
        continue
      if '/' in filename or '\\' in filename:
        logger.warn("object name can't contain '/' or '\\', skipping " + data_file)
        continue
      object_name = os.path.normpath(os.path.join(object_name_prefix,
                                                  relative_dir,
                                                  filename))
      object_name = '/'.join(object_name.split('\\'))
      logger.info('putting %s to %s/%s' % (data_file, bucket_name, object_name))
      put_object(data_file=data_file,
                 bucket_name=bucket_name,
                 object_name=object_name,
                 metadata=metadata)


def put_object(data_file, bucket_name, object_name, metadata, is_archive=False):
  check_bucket_name(bucket_name)
  check_object_name(object_name)
  check_metadata(metadata)
  fds_metadata = parse_metadata_from_str(metadata)
  # The object name doesn't have starting '/'
  object_name = object_name.lstrip('/')
  result = None
  if data_file:
    with open(data_file, "rb") as f:
      result = fds_client.put_object(bucket_name, object_name, f, fds_metadata, is_archive=is_archive)
  else:
    result = fds_client.put_object(bucket_name, object_name, sys.stdin, fds_metadata, is_archive=is_archive)
  logger.debug("Upload object success")

  if result:
    logger.info('Put object %s success' % object_name)
  else:
    logger.info('Put object %s failed' % object_name)


def multipart_upload(bucket_name, object_name, metadata, stream):
  upload_token = None
  try:
    logger.debug('Put object in multipart upload mode')
    upload_token = fds_client.init_multipart_upload(bucket_name=bucket_name,
                                                    object_name=object_name)
    logger.debug('Upload id [' + upload_token.upload_id + ']')
    part_number = 1
    upload_list = []
    while True:
      data = stream.read(multipart_upload_buffer_size)
      if len(data) <= 0:
        break
      logger.info("Part %d read %d bytes" % (part_number, len(data)))

      rtn = None
      for i in range(max_upload_retry_time):
        try:
          rtn = fds_client.upload_part(bucket_name=upload_token.bucket_name,
                                       object_name=upload_token.object_name,
                                       upload_id=upload_token.upload_id,
                                       part_number=part_number,
                                       data=data)
          upload_list.append(rtn)
          break
        except:
          sleepSeconds = (i + 1) * 5
          logger.warning(
            "upload part %d failed, retry after %d seconds" % (part_number, sleepSeconds))
          sleep(sleepSeconds)

      if not rtn:
        raise GalaxyFDSClientException("Upload part %d failed" % part_number)
      part_number += 1

    upload_part_result = UploadPartResultList({"uploadPartResultList": upload_list})
    logger.info("Upload data end, result : %s" % json.dumps(upload_part_result))
    return fds_client.complete_multipart_upload(bucket_name=upload_token.bucket_name,
                                                object_name=upload_token.object_name,
                                                upload_id=upload_token.upload_id,
                                                metadata=metadata,
                                                upload_part_result_list=json.dumps(
                                                  upload_part_result))
    logger.info("Upload complete")
  except Exception as e:
    try:
      logger.error("Upload id %s will be abort" % upload_token.upload_id)
      fds_client.abort_multipart_upload(bucket_name, object_name, upload_token.upload_id)
    except:
      pass
    raise e


def parse_metadata_from_str(metadata):
  fds_metadata = None
  if metadata:
    fds_metadata = FDSObjectMetadata()
    for i in metadata.split(';'):
      key, value = i.split(':', 1)
      if key and value:
        if key.startswith(FDSObjectMetadata.USER_DEFINED_METADATA_PREFIX):
          fds_metadata.add_user_metadata(key, value)
        else:
          fds_metadata.add_header(key, value)
  return fds_metadata


def get_object(data_file, bucket_name, object_name, metadata, offset, length,
    webp_quality=False, gif_extract_type=False, is_archive=False):
  check_bucket_name(bucket_name)
  check_object_name(object_name)
  if webp_quality:
    fds_object = fds_client._get_webp_(bucket_name=bucket_name,
                                         object_name=object_name,
                                         quality=webp_quality)
    length = -1
    offset = 0
  elif gif_extract_type:
    fds_object = fds_client._get_extracted_gif_(bucket_name=bucket_name,
                                                object_name=object_name,
                                                type=gif_extract_type)
    length = -1
    offset = 0
  else:
    fds_object = fds_client.get_object(bucket_name=bucket_name,
                                       object_name=object_name,
                                       position=offset,
                                       stream=True,
                                       is_archive=is_archive)

  length_left = length
  if length_left == -1:
    length_left = IS_PY3 and sys.maxsize or sys.maxint
  try:
    if data_file:
      with open(data_file, "wb") as f:
        for chunk in fds_object.stream:
          # if isinstance(chunk, bytes):
          #   chunk = chunk.decode(encoding='UTF-8')
          l = min(length_left, len(chunk));
          f.write(chunk[0:l])
          length_left -= l
          if length_left <= 0:
            break
    else:
      for chunk in fds_object.stream:
        # if isinstance(chunk, bytes):
        #   chunk = chunk.decode(encoding='UTF-8')
        l = min(length_left, len(chunk))
        if IS_PY3:
          sys.stdout.buffer.write(chunk[0:l])
        else:
          sys.stdout.write(chunk[0:l])
        length_left -= l
        if length_left <= 0:
          break
      sys.stdout.flush()
  finally:
    fds_object.stream.close()


def post_object(data_file, bucket_name, metadata):
  with open(data_file, 'r') as f:
    fds_object = fds_client.post_object(bucket_name=bucket_name, data=f, metadata=metadata)
    logger.debug('Post object [' + fds_object.object_name + ']')
    sys.stdout.write(fds_object.object_name)


def download_directory(bucket_name, object_prefix, data_dir, recursive):
  def mkdirs(path):
    if not os.path.isdir(path):
      os.makedirs(path)

  delimiter = not recursive and '/' or ''

  if not object_prefix.endswith('/'):
    sys.stderr.write('Object prefix must endswith /')
    sys.stderr.flush()
    return

  if not data_dir:
    data_dir = './'

  if not data_dir.endswith('/'):
    data_dir += '/'

  if object_prefix == '/':  # download all objects under bucket
    objects_list = fds_client.list_objects(bucket_name, prefix='', delimiter=delimiter)
  else:
    objects_list = fds_client.list_objects(bucket_name, prefix=object_prefix, delimiter=delimiter)
  # print(objects_list.objects)
  while True:
    for obj in objects_list.objects:
      if obj.object_name.endswith('/'):
        mkdirs(data_dir + obj.object_name)
      else:
        if '/' in obj.object_name:
          mkdirs(data_dir + obj.object_name[:obj.object_name.rfind('/')])
        obj_path = data_dir + obj.object_name
        if os.path.exists(obj_path) and obj.size == os.path.getsize(obj_path):
          logger.debug("[%s/%s] is already downloaded" % (bucket_name, obj.object_name))
        else:
          get_object(data_dir + obj.object_name, bucket_name, obj.object_name, None, 0, -1)
          logger.debug("Download [%s/%s] success" % (bucket_name, obj.object_name))
    for prefix in objects_list.common_prefixes:
      mkdirs(data_dir + prefix)
    if objects_list.is_truncated:
      objects_list = fds_client.list_next_batch_of_objects(objects_list)
    else:
      break

  sys.stdout.write("Downdoad directory[%s] success" % object_prefix)
  sys.stdout.flush()


def put_bucket(bucket_name):
  check_bucket_name(bucket_name=bucket_name)
  fds_client.create_bucket(bucket_name)
  sys.stdout.write("Create bucket[%s] success" % bucket_name)


def get_bucket_acl(bucket_name):
  acl = fds_client.get_bucket_acl(bucket_name=bucket_name)
  sys.stdout.write('ACL:\n')
  sys.stdout.write('gratee_id\tgrant_type\tpermission\n')
  for i in acl.get_grant_list():
    sys.stdout.write(
      str(i.grantee['id']) + '\t' + str(i.type) + '\t' + str(i.permission.to_string()) + '\n')


def put_bucket_acl(bucket_name, gratee_list, permission_list):
  check_bucket_name(bucket_name=bucket_name)
  bucketAcl = AccessControlList()
  for role in gratee_list:
    grant = Grant(Grantee(role), Permission(permission_list).get_value())
    if role in [UserGroups.ALL_USERS, UserGroups.AUTHENTICATED_USERS]:
      grant.type = GrantType.GROUP
    bucketAcl.add_grant(grant)
  fds_client.set_bucket_acl(bucket_name=bucket_name, acl=bucketAcl)


def delete_bucket_acl(bucket_name, gratee_list, permission_list):
  check_bucket_name(bucket_name=bucket_name)
  bucketAcl = AccessControlList()
  for role in gratee_list:
    grant = Grant(Grantee(role), Permission(permission_list).get_value())
    if role in [UserGroups.ALL_USERS, UserGroups.AUTHENTICATED_USERS]:
      grant.type = GrantType.GROUP
    bucketAcl.add_grant(grant)
  fds_client.delete_bucket_acl(bucket_name=bucket_name, acl=bucketAcl)


def delete_object_acl(bucket_name, object_name, gratee_list, permission_list, is_archive=False):
  check_bucket_name(bucket_name)
  check_bucket_name(object_name)
  object_acl = AccessControlList()
  for role in gratee_list:
    grant = Grant(Grantee(role), Permission(permission_list).get_value())
    if role in [UserGroups.ALL_USERS, UserGroups.AUTHENTICATED_USERS]:
      grant.type = GrantType.GROUP
    object_acl.add_grant(grant)
  fds_client.delete_object_acl(bucket_name, object_name, object_acl, is_archive=is_archive)


def get_object_acl(bucket_name, object_name, is_archive=False):
  acl = fds_client.get_object_acl(bucket_name=bucket_name, object_name=object_name, is_archive=is_archive)
  sys.stdout.write('ACL:\n')
  sys.stdout.write('gratee_id\tgrant_type\tpermission\n')
  for i in acl.get_grant_list():
    sys.stdout.write(str(i.grantee['id']) + '\t' + str(i.type) + '\t' + str(i.permission.to_string()) + '\n')


def put_object_acl(bucket_name, object_name, gratee_list, permission_list, is_archive=False):
  check_bucket_name(bucket_name=bucket_name)
  check_object_name(object_name=object_name)
  object_acl = AccessControlList()
  for role in gratee_list:
    grant = Grant(Grantee(role), Permission(permission_list).get_value())
    if role in [UserGroups.ALL_USERS, UserGroups.AUTHENTICATED_USERS]:
      grant.type = GrantType.GROUP
    object_acl.add_grant(grant)
  fds_client.set_object_acl(bucket_name, object_name, object_acl, is_archive=is_archive)
  logger.info('set [%s/%s] acl success', bucket_name, object_name)

def set_public(bucket_name, object_name=None, is_archive=False):
  check_bucket_name(bucket_name)
  fds_client.set_public(bucket_name, object_name=object_name, is_archive=is_archive)
  if object_name:
    if is_archive:
      sys.stdout.write('set archive %s/%s public success' % (bucket_name, object_name))
    else:
      sys.stdout.write('set object %s/%s public success' % (bucket_name, object_name))
  else:
    sys.stdout.write('set bucket %s public success' % bucket_name)
  sys.stdout.flush()

def restore_archived_object(bucket_name, object_name):
  check_bucket_name(bucket_name=bucket_name)
  check_object_name(object_name=object_name)
  fds_client._restore_archived_object_(bucket_name, object_name)
  logger.info('restore archived object [%s/%s] success; this may takes several hours', bucket_name, object_name)

def put_bucket_lifecycle_config(bucket_name, lifecycle):
  check_bucket_name(bucket_name)
  lifecycle = FDSLifecycleConfig(json.loads(lifecycle))
  fds_client.update_lifecycle_config(bucket_name, lifecycle)
  sys.stdout.write("put [%s] lifecycle config success" % bucket_name)
  sys.stdout.flush()


def put_bucket_cors_config(bucket_name, cors):
  check_bucket_name(bucket_name)
  cors = FDSCORSConfig(json.loads(cors))
  fds_client.update_cors_config(bucket_name, cors)
  sys.stdout.write("put [%s] cors config success" % bucket_name)
  sys.stdout.flush()


def put_bucket_lifecycle_rule(bucket_name, lifecycle_rule):
  check_bucket_name(bucket_name)
  rule = FDSLifecycleRule(json.loads(lifecycle_rule))
  fds_client.update_lifecycle_rule(bucket_name, rule)
  sys.stdout.write("put [%s] lifecycle rule success" % bucket_name)
  sys.stdout.flush()


def put_bucket_cors_rule(bucket_name, cors_rule):
  check_bucket_name(bucket_name)
  rule = FDSCORSRule(json.loads(cors_rule))
  fds_client.update_cors_rule(bucket_name, rule)
  sys.stdout.write("put [%s] cors rule success" % bucket_name)
  sys.stdout.flush()


def get_bucket_lifecycle_config(bucket_name):
  lifecycle = fds_client.get_lifecycle_config(bucket_name);
  sys.stdout.write(json.dumps(lifecycle))
  sys.stdout.flush()


def get_bucket_cors_config(bucket_name):
  cors = fds_client.get_cors_config(bucket_name)
  sys.stdout.write(json.dumps(cors))
  sys.stdout.flush()


def delete_object(bucket_name, object_name, **kwargs):
  check_bucket_name(bucket_name=bucket_name)
  check_object_name(object_name=object_name)
  fds_client.delete_object(bucket_name=bucket_name,
                           object_name=object_name,
                           **kwargs)
  logger.info('delete object %s success' % (object_name))


def delete_objects(bucket_name, **kwargs):
  check_bucket_name(bucket_name=bucket_name)
  for obj in fds_client.list_all_objects(bucket_name, prefix=kwargs["object_prefix"], delimiter=""):
    fds_client.delete_object(bucket_name=bucket_name,
                             object_name=obj.object_name,
                             **kwargs)
    logger.info('delete object %s success' % (obj.object_name))


def delete_bucket(bucket_name):
  if fds_client.does_bucket_exist(bucket_name):
    fds_client.delete_bucket(bucket_name=bucket_name)
    logger.info('delete bucket success')
  else:
    logger.info('bucket does not exist')


def delete_bucket_and_objects(bucket_name):
  if fds_client.does_bucket_exist(bucket_name):
    for obj in fds_client.list_all_objects(bucket_name, prefix="", delimiter=""):
      fds_client.delete_object(bucket_name, obj.object_name)
    for obj in fds_client.list_all_objects(bucket_name, prefix="", delimiter="", is_archive=True):
      fds_client.delete_object(bucket_name, obj.object_name, is_archive=True)
    fds_client.delete_bucket(bucket_name)
    logger.info('delete bucket and objects success')
  else:
    logger.info('bucket does not exist')


def head_object(bucket_name, object_name):
  res = fds_client.does_object_exists(bucket_name=bucket_name,
                                      object_name=object_name)
  if res:
    logger.info('%s exists' % object_name)
  else:
    logger.info('%s does not exists' % object_name)


def head_bucket(bucket_name):
  return fds_client.does_bucket_exist(bucket_name=bucket_name)


def list_directory(bucket_name, object_name_prefix, start_mark):
  if not object_name_prefix:
    object_name_prefix = ''
  path_prefix = object_name_prefix
  if len(path_prefix) > 0 and not path_prefix.endswith('/'):
    path_prefix = path_prefix + '/'
  list_result = fds_client.list_objects(bucket_name=bucket_name,
                                        prefix=path_prefix,
                                        delimiter='/',
                                        )
  if start_mark:
    logger.info('start_marker: ' + start_mark)
    list_result.next_marker = bucket_name + '/' + path_prefix + start_mark
    list_result.is_truncated = True
    list_result = fds_client.list_next_batch_of_objects(list_result)

  prefix_len = len(path_prefix)

  while True:
    for i in list_result.common_prefixes:
      sys.stdout.write(i[prefix_len:])
      sys.stdout.write('\n')
    for i in list_result.objects:
      sys.stdout.write(i.object_name[prefix_len:])
      sys.stdout.write('\n')
    sys.stdout.flush()
    if not list_result.is_truncated:
      break
    list_result = fds_client.list_next_batch_of_objects(list_result)

def list_bucket_trash(bucket_name=None):
  prefix = bucket_name and bucket_name + "/" or ""

  list_result = fds_client.list_trash_objects(prefix=prefix, delimiter="")

  for i in list_result.common_prefixes:
    sys.stdout.write(i)
    sys.stdout.write('\n')
  for i in list_result.objects:
    sys.stdout.write(i.object_name)
    sys.stdout.write('\n')
  sys.stdout.flush()
  if list_result.is_truncated:
    sys.stdout.write('...\n')

def list_object(bucket_name, object_name_prefix, start_mark='', is_archive=False):
  list_result = fds_client.list_objects(bucket_name=bucket_name,
                                        prefix=object_name_prefix,
                                        delimiter='',
                                        is_archive=is_archive)
  if start_mark:
    list_result.is_truncated = True
    list_result.next_marker = bucket_name + '/' + object_name_prefix + start_mark
    list_result = fds_client.list_next_batch_of_objects(list_result)

  for i in list_result.common_prefixes:
    sys.stdout.write(i)
    sys.stdout.write('\n')
  for i in list_result.objects:
    sys.stdout.write(i.object_name)
    sys.stdout.write('\n')
  sys.stdout.flush()
  if list_result.is_truncated:
    sys.stdout.write('...\n')


def list_version_ids(bucket_name, object_name):
  vids = fds_client._list_version_ids_(bucket_name, object_name)
  sys.stdout.write(json.dumps(vids))
  sys.stdout.flush()


def copy_object(src_bucket_name, src_object_name, dst_bucket_name, dst_object_name):
  check_bucket_name(src_bucket_name)
  check_bucket_name(dst_bucket_name)
  check_object_name(src_object_name)
  check_object_name(dst_object_name)

  result = None
  result = fds_client.copy_object(src_bucket_name, src_object_name, dst_bucket_name,
                                  dst_object_name)

  logger.debug("Copy object success")

  if result:
    logger.info('Successfully copy object %s from bucket %s to bucket %s, and rename it to %s' % (
      src_object_name, src_bucket_name, dst_bucket_name, dst_object_name))
  else:
    logger.info('Failed to copy object %s from bucket %s to bucket %s, and rename it to %s' % (
      src_object_name, src_bucket_name, dst_bucket_name, dst_object_name))


def set_bucket_default_webp_quality(bucket_name, webp_quality):
  fds_client._set_bucket_default_webp_quality_(bucket_name, webp_quality)
  res = fds_client._get_bucket_default_webp_quality_(bucket_name)
  if res > 0:
    logger.info("[%s] default webp quality: %d" % (bucket_name, res))
  else:
    logger.info("[%s] is not set convert webp" % bucket_name)


def set_bucket_default_gif_extract_type(bucket_name, gif_extract_type):
  fds_client._set_bucket_default_gif_extract_type_(bucket_name, gif_extract_type)
  res = fds_client._get_bucket_default_gif_extract_type_(bucket_name)
  if res == 'unknown':
    logger.info("[%s] is not set auto gif extract" % (bucket_name))
  else:
    logger.info("[%s] default gif extract type: %s" % (bucket_name, res))


def main():
  parser = argparse.ArgumentParser(description=__doc__,
                                   formatter_class=argparse.RawDescriptionHelpFormatter,
                                   epilog="Doc - http://docs.api.xiaomi.com/fds/")

  parser.add_argument('-m', '--method',
                      nargs='?',
                      metavar='method',
                      const='put',
                      type=str,
                      dest='method',
                      help='Method of the request. Can be one of put/get/delete/post/head (default: put)'
                      ).completer = ChoicesCompleter(('put', 'get', 'delete', 'post', 'head'))

  parser.add_argument('-b', '--bucket',
                      nargs='?',
                      metavar='bucket',
                      type=str,
                      dest='bucket',
                      help='Name of bucket to operate'
                      ).completer = bucket_name_completer

  parser.add_argument('-o', '--object',
                      nargs='?',
                      metavar='object',
                      type=str,
                      dest='object',
                      help='Name of object to operate'
                      )

  parser.add_argument('-r', '--region',
                      nargs='?',
                      metavar='region',
                      type=str,
                      dest='region',
                      help='Can be one of cnbj0/cnbj1/cnbj2/awsbj0/awsusor0/awssgp0/awsde0 (default: cnbj0)'
                      )

  parser.add_argument('-e', '--end_point',
                      nargs='?',
                      metavar='end point',
                      type=str,
                      dest='end_point',
                      help='can be [cnbj1.fds.api.xiaomi.com] or empty'
                      )

  parser.add_argument('-c', '--CDN',
                      metavar='CDN',
                      action='store_const',
                      const=False,
                      dest='CDN',
                      default=False,
                      help='If toggled, CDN is enabled'
                      )

  parser.add_argument('-p', '--presigned_url',
                      action='store_true',
                      dest='presigned_url',
                      default=False,
                      help='If toggled, generate presigned url'
                      )

  parser.add_argument('--https',
                      metavar='https',
                      nargs='?',
                      dest='https',
                      default=True,
                      help='If toggled, https is enabled'
                      )

  parser.add_argument('--ak',
                      nargs='?',
                      metavar='ACCESS_KEY',
                      dest='ak',
                      help='Specify access key'
                      )

  parser.add_argument('--sk',
                      nargs='?',
                      metavar='SECRET_KEY',
                      dest='sk',
                      help='Specify secret key'
                      )

  parser.add_argument('-L', '--list',
                      nargs='?',
                      metavar='list directory',
                      const='',
                      type=str,
                      dest='list_dir',
                      help='List Bucket/Object under current user')

  parser.add_argument('-l', '--list_objects',
                      nargs='?',
                      metavar='list objects',
                      const='',
                      type=str,
                      dest='list_objects',
                      help='List Bucket/Object under current user')

  parser.add_argument('--list-trash',
                      action='store_true',
                      dest='list_trash',
                      default=False,
                      help='''If toggled, List trash under current user.''')

  parser.add_argument('-d', '--data',
                      nargs='?',
                      metavar='data file',
                      dest='data_file',
                      help='file to be uploaded or stored')

  parser.add_argument('-D', '--directory',
                      nargs='?',
                      metavar='data dir',
                      dest='data_dir',
                      help="Directory to be uploaded or stored. Use '/' to download all objects under the bucket")

  parser.add_argument('-R', '--recursive',
                      action='store_true',
                      dest='recursive',
                      default=False,
                      help='If toggled, download the directory recursively')

  parser.add_argument('--offset',
                      nargs='?',
                      metavar='offset',
                      type=int,
                      const=0,
                      default=0,
                      dest='offset',
                      help='offset of object to be read')

  parser.add_argument('--length',
                      nargs='?',
                      metavar='length',
                      type=int,
                      dest='length',
                      const=-1,
                      default=-1,
                      help='length of object to be read')

  parser.add_argument('--metadata',
                      nargs='?',
                      metavar='meta data of object to be uploaded',
                      dest='metadata',
                      help='example: "content-type:text/json;x-xiaomi-meta-user-defined:foo"')

  parser.add_argument('--start',
                      nargs='?',
                      metavar='start mark',
                      type=str,
                      dest='start_mark',
                      const=None,
                      default=None,
                      help='used with -l or -L option, returned object name should be *no less* than start mark in dictionary order'
                      )

  parser.add_argument('--debug',
                      metavar='debug',
                      action='store_const',
                      const=True,
                      default=False,
                      dest='debug',
                      help='If toggled, print debug log')

  parser.add_argument('--expiration',
                      nargs='?',
                      type=str,
                      metavar='expiration in hour',
                      default='1.0',
                      dest='expiration_in_hour',
                      help='used with --presigned_url, set expiration of presigned url generated from now on(hour), default to one hour')

  parser.add_argument('--force',
                      action='store_true',
                      dest='force_delete',
                      default=False,
                      help='If toggled, delete bucket and objects')

  parser.add_argument('--disable_trash',
                      action='store_true',
                      dest='disable_trash',
                      default=False,
                      help='If toggled, delete object without move to trash')

  parser.add_argument('-P', '--object_prefix',
                      nargs='?',
                      metavar="object's prefix",
                      type=str,
                      dest='object_prefix',
                      help="object's prefix")

  parser.add_argument('--lifecycle',
                      nargs='?',
                      metavar='lifecycle config, json format',
                      dest='lifecycle',
                      const=True,
                      default=None,
                      help='''Put or get lifecycle configof the bucket. Please use \\" instead of " in this argument when putting lifecycle config due to shell may eat double quotes.''')

  parser.add_argument('--cors',
                      nargs='?',
                      metavar='cors config, json format',
                      dest='cors',
                      const=True,
                      default=None,
                      help='''Put or get cors config of the bucket. Please use \\" instead of " in this argument when putting cors config due to shell may eat double quotes.''')

  parser.add_argument('--lifecycle-rule',
                      nargs='?',
                      metavar='lifecycle rule, json format',
                      dest='lifecycle_rule',
                      const=True,
                      default=None,
                      help='''Add/update or get one rule of lifecycle config of the bucket. Please use \\" instead of " in this argument when putting lifecycle config due to shell may eat double quotes.''')

  parser.add_argument('--cors-rule',
                      nargs='?',
                      metavar='cors rule, json format',
                      dest='cors_rule',
                      const=True,
                      default=None,
                      help='''Add/update or get one rule of cors config of the bucket. Please use \\" instead of " in this argument when putting cors config due to shell may eat double quotes.''')


  parser.add_argument('--webp-quality',
                       nargs='?',
                       dest='webp_quality',
                       const=-1,
                       default=None,
                       help='Integer indicates webp quality, -1 will disable bucket auto convert webp')

  parser.add_argument('--gif-extract-type',
                      nargs='?',
                      dest='gif_extract_type',
                      const='unknown',
                      default=None,
                      help='String indicates gif extract type, unknown will disable bucket auto gif extract')

  parser.add_argument('--archive',
                      action='store_true',
                      dest='is_archive',
                      default=False,
                      help='''If toggled, visit archived objects.''')

  parser.add_argument('--restore-archive',
                      action='store_true',
                      dest='restore_archive',
                      default=False,
                      help='''If toggled, restore archived object.''')

  group = parser.add_argument_group('acl')
  group.add_argument('--gratee',
                     nargs='+',
                     metavar='user, group, ALL_USERS, AUTHENTICATED_USERS',
                     dest='gratee',
                     help='Add acl to bucket')
  group.add_argument('--permission',
                     nargs='?',
                     metavar="READ, WRITE, READ_OBJECTS, FULL_CONTROL",
                     dest='permission',
                     choices=['READ', 'WRITE', 'READ_OBJECTS', 'FULL_CONTROL'],
                     help='Add acl to bucket')
  group.add_argument('--public',
                     action='store_true',
                     dest='is_public',
                     default=False,
                     help='''If toggled, set bucket or object public.''')

  cp = parser.add_argument_group('cp')
  cp.add_argument('-srcb', '--src_bucket',
                  nargs='?',
                  metavar='bucket, name',
                  dest='src_bucket_name',
                  help='Copy object from src_bucket to dst_bucket')
  cp.add_argument('-srco', '--src_object',
                  nargs='?',
                  metavar='object, name',
                  dest='src_object_name',
                  help='Copy object from src_bucket to dst_bucket')
  cp.add_argument('-dstb', '--dst_bucket',
                  nargs='?',
                  metavar='bucket, name',
                  dest='dst_bucket_name',
                  help='Copy object from src_bucket to dst_bucket')
  cp.add_argument('-dsto', '--dst_object',
                  nargs='?',
                  metavar='object, name',
                  dest='dst_object_name',
                  help='Copy object from src_bucket to dst_bucket')

  argcomplete.autocomplete(parser)

  args = parser.parse_args()

  # set logging
  log_format = '%(asctime)-15s [%(filename)s:%(lineno)d] %(message)s'
  logging.basicConfig(format=log_format)
  global logger
  logger = logging.getLogger('fds.cmd')
  debug_enabled = args.debug

  if debug_enabled:
    logger.setLevel(logging.DEBUG)
  else:
    logger.setLevel(logging.INFO)
  ## read config
  parse_argument(args=args)

  check_region(region=region)
  check_bucket_name(bucket_name=bucket_name)
  global fds_config
  fds_config = FDSClientConfiguration(region_name=region,
                                      enable_https=enable_https,
                                      enable_cdn_for_download=enable_cdn,
                                      enable_cdn_for_upload=enable_cdn,
                                      threshold_size=multipart_upload_threshold_size,
                                      part_size=multipart_upload_buffer_size)
  global end_point
  if not end_point is None:
    fds_config.set_endpoint(end_point)
  global fds_client
  fds_client = GalaxyFDSClient(access_key=access_key,
                               access_secret=secret_key,
                               config=fds_config)

  global force_delete
  global disable_trash
  global object_prefix
  global recursive, webp_quality, gif_extract_type

  try:
    if presigned_url:
      expiration = int(1000 * (float(eval(expiration_in_hour)) * 3600 +
                               float((datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0,
                                                                   0)).total_seconds())))
      meta = parse_metadata_from_str(metadata=metadata)
      content_type = None
      if meta and 'content-type' in meta.metadata:
        content_type = meta.metadata['content-type']
      url = fds_client.generate_presigned_uri(fds_config.get_base_uri(),
                                              bucket_name=bucket_name, object_name=object_name,
                                              expiration=expiration, http_method=method.upper(),
                                              content_type=content_type)
      print(url)
    elif not (list_dir is None):
      if not (bucket_name is None):
        list_directory(bucket_name=bucket_name,
                       object_name_prefix=list_dir, start_mark=start_mark)
      else:
        list_buckets(fds_client=fds_client, prefix=list_dir, start_mark=start_mark)
    elif list_trash:
      list_bucket_trash(bucket_name=bucket_name)
    elif not (list_objects is None):
      if not (bucket_name is None):
        if object_name is not None:
          list_version_ids(bucket_name=bucket_name, object_name=object_name)
        else:
          list_object(bucket_name=bucket_name, object_name_prefix=list_objects,
                      start_mark=start_mark, is_archive=is_archive)
      else:
        list_buckets(fds_client=fds_client, prefix=list_objects, start_mark=start_mark)
      pass
    else:
      if method == 'put':
        if src_bucket_name and src_object_name and dst_bucket_name and dst_object_name:
          copy_object(src_bucket_name, src_object_name, dst_bucket_name, dst_object_name)
        elif object_name:
          if data_dir:
            put_directory(data_dir=data_dir,
                          bucket_name=bucket_name,
                          object_name_prefix=object_name,
                          metadata=metadata)
          elif is_public:
            set_public(bucket_name=bucket_name, object_name=object_name, is_archive=is_archive)
          elif gratee and permission:
            put_object_acl(bucket_name, object_name, gratee, permission, is_archive)
          elif restore_archive:
            restore_archived_object(bucket_name, object_name)
          else:
            put_object(data_file=data_file,
                       bucket_name=bucket_name,
                       object_name=object_name,
                       metadata=metadata,
                       is_archive=is_archive)
        elif is_public:
          set_public(bucket_name=bucket_name)
        elif gratee and permission:
          put_bucket_acl(bucket_name, gratee, permission)
        elif lifecycle:
          put_bucket_lifecycle_config(bucket_name, lifecycle)
        elif lifecycle_rule:
          put_bucket_lifecycle_rule(bucket_name, lifecycle_rule)
        elif cors:
          put_bucket_cors_config(bucket_name, cors)
        elif cors_rule:
          put_bucket_cors_rule(bucket_name, cors_rule)
        elif webp_quality:
          set_bucket_default_webp_quality(bucket_name=bucket_name, webp_quality=webp_quality)
        elif gif_extract_type:
          set_bucket_default_gif_extract_type(bucket_name=bucket_name, gif_extract_type=gif_extract_type)
        else:
          put_bucket(bucket_name)
        pass
      elif method == 'get':
        if object_prefix:
          download_directory(bucket_name=bucket_name, object_prefix=object_prefix,
                             data_dir=data_dir, recursive=recursive)
        elif object_name:
          get_object(data_file=data_file,
                     bucket_name=bucket_name,
                     object_name=object_name,
                     metadata=metadata,
                     offset=offset,
                     length=length,
                     webp_quality=webp_quality,
                     gif_extract_type=gif_extract_type,
                     is_archive=is_archive)
        elif lifecycle:
          get_bucket_lifecycle_config(bucket_name)
        elif cors:
          get_bucket_cors_config(bucket_name)
        else:
          get_bucket_acl(bucket_name=bucket_name)
        pass
      elif method == 'post':
        post_object(data_file=data_file,
                    bucket_name=bucket_name,
                    metadata=metadata)
        pass
      elif method == 'delete':
        if object_name:
          if gratee and permission:
            delete_object_acl(bucket_name, object_name, gratee, permission, is_archive)
          else:
            delete_object(bucket_name=bucket_name,
                          object_name=object_name,
                          enable_trash=not disable_trash,
                          is_archive=is_archive)
        elif object_prefix is not None:
          delete_objects(bucket_name=bucket_name,
                         object_prefix=object_prefix,
                         enable_trash=not disable_trash)
        elif force_delete:
          delete_bucket_and_objects(bucket_name=bucket_name)
        else:
          if gratee and permission:
            delete_bucket_acl(bucket_name, gratee, permission)
          else:
            delete_bucket(bucket_name=bucket_name)
        pass
      elif method == 'head':
        if object_name:
          if not head_object(bucket_name=bucket_name,
                             object_name=object_name):
            exit(1)
        else:
          if not head_bucket(bucket_name=bucket_name):
            exit(1)
      else:
        parser.print_help()

  except Exception as e:
    print(e)
    print("\n")

    ex_type, ex, tb = sys.exc_info()
    traceback.print_tb(tb)

    # sys.stderr.write(str(e))
    # sys.stderr.flush()
    if debug_enabled:
      logger.debug(e, exc_info=True)
    exit(1)


if __name__ == "__main__":
  main()



class AccessLogConfig(dict):
  '''
  The Access Log Config Class:
  '''

  def __init__(self, json):
    if 'bucketName' in json.keys():
      self.bucketName = json['bucketName']
    if 'enabled' in json.keys():
      self.enabled = json['enabled']
    if 'logBucketName' in json.keys():
      self.logBucketName = json['logBucketName']
    if 'logPrefix' in json.keys():
      self.logPrefix = json['logPrefix']
    if 'cdnEnabled' in json.keys():
      self.cdnEnabled = json['cdnEnabled']
    if 'cdnLogPrefix' in json.keys():
      self.cdnLogPrefix = json['cdnLogPrefix']

  @property
  def cdnEnabled(self):
    return self.cdnEnabled

  @cdnEnabled.setter
  def cdnEnabled(self, cdnEnabled):
    self['cdnEnabled'] = cdnEnabled

  @property
  def logPrefix(self):
    return self.logPrefix

  @logPrefix.setter
  def logPrefix(self, logPrefix):
    self['logPrefix'] = logPrefix

  @property
  def enabled(self):
    return self.enabled

  @enabled.setter
  def enabled(self, enabled):
    self['enabled'] = enabled

  @property
  def logBucketName(self):
    return self.logBucketName

  @logBucketName.setter
  def logBucketName(self, logBucketName):
    self['logBucketName'] = logBucketName

  @property
  def bucketName(self):
    return self.bucketName

  @bucketName.setter
  def bucketName(self, bucketName):
    self['bucketName'] = bucketName

  @property
  def cdnLogPrefix(self):
    return self.cdnLogPrefix

  @cdnLogPrefix.setter
  def cdnLogPrefix(self, cdnLogPrefix):
    self['cdnLogPrefix'] = cdnLogPrefix



# coding=utf-8
import sys
import unittest

sys.path.append('../')

from multiprocessing.pool import ThreadPool
from fds import GalaxyFDSClient
from fds.fds_client_configuration import FDSClientConfiguration
from test.test_common import region_name, access_key, access_secret


class multithreadingClientTest(unittest.TestCase):
  def setUp(self):
    config = FDSClientConfiguration(region_name, False, False, False)
    self.client = GalaxyFDSClient(access_key, access_secret, config)
    self.bucket_name = "1024"

  def checkAndDeleteBucket(self, index):
    for i in range(50):
      self.client.list_objects(self.bucket_name, "", "/")
    print
    index

  def testMultithreadingClient(self):
    pool = ThreadPool(50)
    if not self.client.does_bucket_exist(self.bucket_name):
      self.client.create_bucket(self.bucket_name)
    pool.map(self.checkAndDeleteBucket, range(50))
    pool.close()
    pool.join()



from fds.model.put_object_result import PutObjectResult


class CopyObjectResult(PutObjectResult):
  def from_put_object_result(self, put_object_result):
    self.access_key_id = put_object_result.access_key_id
    self.bucket_name = put_object_result.bucket_name
    self.expires = put_object_result.expires
    self.object_name = put_object_result.object_name
    self.previous_version_id = put_object_result.previous_version_id
    self.signature = put_object_result.signature
    return self



import sys
import unittest

sys.path.append('../')
from fds.auth import Common
from fds.auth.signature import Signer


class SignerTest(unittest.TestCase):
  signer = Signer('', '')

  def test_get_expires(self):
    uri = '/fds/mybucket/photos/puppy.jpg?test&GalaxyAccessKeyId=' \
          'AKIAIOSFODNN7EXAMPLE&Expires=1141889120&Signature=vjby' \
          'PxybdZaNmGa%2ByT272YEAiv4%3D'
    self.assertEqual(1141889120, self.signer._get_expires(uri))

  def test_header_value(self):
    header = {'Content-Type': 'application/json',
              'x-xiaomi-xx': ['A', 'B', 'C']}
    self.assertEquals('application/json',
                      self.signer._get_header_value(header, 'Content-Type'))
    self.assertEquals('A', self.signer._get_header_value(header, 'x-xiaomi-xx'))
    self.assertEquals('', self.signer._get_header_value(header, 'Content-MD5'))

  def test_canonical_header(self):
    headers = None
    self.assertEqual('', self.signer._canonicalize_xiaomi_headers(headers))
    headers = {}
    headers['Content-Type'] = 'application/json'
    headers[Common.XIAOMI_HEADER_PREFIX + 'meta-username'] = ['x@xiaomi.com ',
                                                              'a@xiaomi.com ']
    headers[
      Common.XIAOMI_HEADER_PREFIX + 'date'] = 'Tue, 27 Mar 2007 21:20:26+000'
    self.assertEqual(
      Common.XIAOMI_HEADER_PREFIX + 'date:' + 'Tue, 27 Mar 2007 21:20:26+000\n'
      + Common.XIAOMI_HEADER_PREFIX + 'meta-username:x@xiaomi.com,a@xiaomi.com\n',
      self.signer._canonicalize_xiaomi_headers(headers))

  def test_canonical_resource(self):
    uri = '/fds/mybucket/?acl&a=1&b=2&c=3'
    canonicalized_resource = self.signer._canonicalize_resource(uri)
    self.assertEqual('/fds/mybucket/?acl', canonicalized_resource)

    uri = '/fds/mybucket/test.txt?uploads&uploadId=xxx&partNumber=3&' \
          'timestamp=12345566';
    canonicalized_resource = self.signer._canonicalize_resource(uri)
    self.assertEqual(
      '/fds/mybucket/test.txt?partNumber=3&uploadId=xxx&uploads',
      canonicalized_resource)

  def test_construct_string_to_sign(self):
    http_method = 'GET'
    headers = None
    uri = '/fds/bucket/test.txt?uploads&uploadId=xx&partNumber=1'

    self.assertEquals('%s\n\n\n\n' \
                      '/fds/bucket/test.txt?partNumber=1&uploadId=xx&uploads' %
                      http_method,
                      self.signer._construct_string_to_sign(http_method, headers, uri))

    headers = {}
    headers[Common.CONTENT_TYPE] = 'application/json'
    headers[Common.CONTENT_MD5] = '123131331313231'
    headers[Common.DATE] = 'Tue, 27 Mar 2007 21:20:26+0000'

    self.assertEquals(
      '%s\n%s\n%s\n%s\n''/fds/bucket/test.txt?partNumber=1&uploadId=xx&uploads' % (
        http_method, headers[Common.CONTENT_MD5], headers[Common.CONTENT_TYPE],
        headers[Common.DATE]),
      self.signer._construct_string_to_sign(http_method, headers, uri))

    headers['%sdate' % Common.XIAOMI_HEADER_PREFIX] = \
      'Tue, 28 Mar 2007 21:20:26+0000'
    self.assertEquals(
      '%s\n%s\n%s\n\n%sdate:Tue, 28 Mar 2007 21:20:26+0000\n' \
      '/fds/bucket/test.txt?partNumber=1&uploadId=xx&uploads' % (
        http_method,
        headers[Common.CONTENT_MD5],
        headers[Common.CONTENT_TYPE],
        Common.XIAOMI_HEADER_PREFIX),
      self.signer._construct_string_to_sign(http_method, headers, uri))

    uri = '/fds/bucket/test.txt?GalaxyAccessKeyId=AKIAIOSFODNN7EXAMPLE' \
          '&Expires=1141889120&Signature=vjbyPxybdZaNmGa%2ByT272YEAiv4%3D'
    self.assertEquals(
      '%s\n%s\n%s\n1141889120\n%sdate:Tue, 28 Mar 2007 21:20:26+0000\n' \
      '/fds/bucket/test.txt' % (
        http_method,
        headers[Common.CONTENT_MD5],
        headers[Common.CONTENT_TYPE],
        Common.XIAOMI_HEADER_PREFIX),
      self.signer._construct_string_to_sign(http_method, headers, uri))



# coding=utf-8
from __future__ import print_function

import unittest
from sys import version_info

import time

IS_PY3 = version_info[0] >= 3

if IS_PY3:
  from urllib.request import urlopen
else:
  from urllib2 import urlopen

import hashlib

import sys

sys.path.append('../')
import os
from os.path import expanduser

from fds.auth import Common
from fds.galaxy_fds_client import GalaxyFDSClient
from fds.fds_client_configuration import FDSClientConfiguration
from fds.model.fds_object_metadata import FDSObjectMetadata
from fds.model.permission import Permission
from fds.model.permission import AccessControlList
from fds.model.permission import Grant
from fds.model.permission import Grantee
from fds.model.upload_part_result_list import UploadPartResultList
from fds.model.fds_lifecycle import *
from fds.model.fds_cors import FDSCORSConfig,FDSCORSRule
import json

from datetime import datetime


class ClientTest(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    ClientTest.init_from_local_config()
    config = FDSClientConfiguration(
      region_name=region_name,
      enable_https=False,
      enable_cdn_for_upload=False,
      enable_cdn_for_download=False,
      endpoint=endpoint)
    config.enable_md5_calculate = True
    ClientTest.client = GalaxyFDSClient(access_key, access_secret, config)
    ClientTest.bucket_name = 'test-python-' + datetime.strftime(datetime.now(), "%Y%m%d%H%M%S%z")
    ClientTest.delete_objects_and_bucket(cls.client, cls.bucket_name)
    ClientTest.client.create_bucket(cls.bucket_name)

  @classmethod
  def tearDownClass(cls):
    ClientTest.delete_objects_and_bucket(ClientTest.client, ClientTest.bucket_name)

  @staticmethod
  def delete_objects_and_bucket(client, bucket_name):
    if client.does_bucket_exist(bucket_name):
      for obj in client.list_all_objects(bucket_name, delimiter=""):
        client.delete_object(bucket_name, obj.object_name)
      client.delete_bucket(bucket_name)

  @staticmethod
  def init_from_local_config():
    global access_key, access_secret, endpoint, app_id, acl_ak, acl_access_secret, region_name
    # if type(access_key) == str and access_key.strip() != "":
    #   return
    config_dirs = [os.path.join(expanduser("~"), ".config", "xiaomi", "config"),
                   os.path.join(expanduser("~"), ".config", "fds", "client.config")];
    config = {}
    for config_dir in config_dirs:
      if not os.path.exists(config_dir):
        pass
      else:
        with open(config_dir) as f:
          config = json.load(fp=f)
    access_key = config.get("xiaomi_access_key_id", "")
    access_secret = config.get("xiaomi_secret_access_key", "")
    endpoint = config.get("xiaomi_fds_endpoint", "")
    app_id = config.get("xiaomi_app_id", "")
    acl_ak = config.get("xiaomi_acl_access_key", "")
    acl_access_secret = config.get("xiaomi_acl_access_secret", "")
    region_name = config.get("xiaomi_region_name", "")

  def test_set_endpoint(self):
    httpConfig = FDSClientConfiguration(region_name, False, False, False)
    endpoint = region_name + ".api.xiaomi.net"
    httpConfig.set_endpoint(endpoint)
    httpUri = httpConfig.get_base_uri()
    self.assertEqual("http://" + endpoint + "/", httpUri)
    httpsConfig = FDSClientConfiguration(region_name, False, False, True)
    httpsConfig.set_endpoint(endpoint)
    httpsUri = httpsConfig.get_base_uri()
    self.assertEqual("https://" + endpoint + "/", httpsUri)

  def test_uri(self):
    client = GalaxyFDSClient(access_key, access_secret,
                             FDSClientConfiguration(region_name, False, False, False))
    bucket_name = self.bucket_name + "1"
    if (client.does_bucket_exist(bucket_name)):
      client.delete_bucket(bucket_name)
    client.create_bucket(bucket_name)
    client.delete_bucket(bucket_name)

  def test_normal_bucket(self):
    for bucket in self.client.list_buckets():
      print(bucket)
    bucket_name = self.bucket_name + "2"
    self.assertFalse(self.client.does_bucket_exist(bucket_name))
    self.client.create_bucket(bucket_name)
    self.assertTrue(self.client.does_bucket_exist(bucket_name))
    self.client.delete_bucket(bucket_name)
    self.assertFalse(self.client.does_bucket_exist(bucket_name))

  def test_create_and_delete_bucket(self):
    print(self.client.list_buckets())
    bucket_name = self.bucket_name + "3"
    try:
      self.client.create_bucket(bucket_name)
    except GalaxyFDSClientException as e:
      print(e.message)
    self.delete_objects_and_bucket(self.client, bucket_name)
    self.client.create_bucket(bucket_name)
    self.assertEquals(True, self.client.does_bucket_exist(bucket_name))
    self.client.delete_bucket(bucket_name)
    self.assertEquals(False, self.client.does_bucket_exist(bucket_name))
    try:
      self.client.delete_bucket(bucket_name)
    except GalaxyFDSClientException as e:
      print(e.message)

  def test_normal_object(self):
    object_name = "testPutGetObject_name"
    self.client.put_object(self.bucket_name, object_name, '')
    self.assertTrue(
      self.client.does_object_exists(self.bucket_name, object_name))
    print(self.client.list_objects(self.bucket_name))
    self.client.delete_object(self.bucket_name, object_name)
    self.assertFalse(
      self.client.does_object_exists(self.bucket_name, object_name))
    print(self.client.list_objects(self.bucket_name))

  def test_bucket_acl(self):
    print(self.bucket_name)

    self.client.get_bucket_acl(self.bucket_name)
    bucketAcl = AccessControlList()
    bucketAcl.add_grant(Grant(Grantee("111"), Permission.READ))
    bucketAcl.add_grant(Grant(Grantee('109901'), Permission.FULL_CONTROL))
    bucketAcl.add_grant(Grant(Grantee('123456'), Permission.SSO_WRITE))
    bucketAcl.add_grant(Grant(Grantee(app_id), Permission.FULL_CONTROL))
    self.client.set_bucket_acl(self.bucket_name, bucketAcl)

    aclListGot = self.client.get_bucket_acl(self.bucket_name)
    readAclCnt = 0
    fullControlCnt = 0
    writeWithSSOCnt = 0
    for i in aclListGot.get_grant_list():
      if i['grantee']['id'] == '111':
        self.assertTrue(i['permission'].to_string() == Permission(Permission.READ).to_string())
        readAclCnt += 1
      elif i['grantee']['id'] == '109901':
        self.assertTrue(
          i['permission'].to_string() == Permission(Permission.FULL_CONTROL).to_string())
        fullControlCnt += 1
      elif i['grantee']['id'] == '123456':
        self.assertTrue(i['permission'].to_string() == Permission(Permission.SSO_WRITE).to_string())
        writeWithSSOCnt += 1
    self.assertTrue(readAclCnt == 1)
    self.assertTrue(fullControlCnt == 1)
    self.assertTrue(writeWithSSOCnt == 1)

    #    self.client.set_bucket_acl(self.bucket_name, bucketAcl)
    acl = self.client.get_bucket_acl(self.bucket_name)
    self.assertTrue(bucketAcl.is_subset(acl))
    acl_client = GalaxyFDSClient(acl_ak, acl_access_secret,
                                 FDSClientConfiguration(region_name, False, False, False,
                                                        endpoint=endpoint))
    object_name = "testBucketAcl7"
    acl_client.put_object(self.bucket_name, object_name, "hahhah")
    self.assertTrue(
      self.client.does_object_exists(self.bucket_name, object_name))
    acl_client.list_objects(self.bucket_name)
    acl_client.delete_object(self.bucket_name, object_name)
    self.assertFalse(
      self.client.does_object_exists(self.bucket_name, object_name))
    self.assertTrue(acl_client.does_bucket_exist(self.bucket_name))

  #    acl_client.delete_bucket(self.bucket_name)
  #    self.assertFalse(self.client.does_bucket_exist(self.bucket_name))

  def test_object_acl(self):
    object_name = "test1"
    content = "test1"
    self.client.put_object(self.bucket_name, object_name, content)
    for bucket in self.client.list_objects(self.bucket_name):
      print(bucket)
    print(self.client.get_object_acl(self.bucket_name, object_name))
    objectAcl = AccessControlList()
    objectAcl.add_grant(Grant(Grantee("111"), Permission.READ))
    objectAcl.add_grant(Grant(Grantee("109901"), Permission.FULL_CONTROL))
    objectAcl.add_grant(Grant(Grantee(acl_ak), Permission.FULL_CONTROL))
    self.client.set_object_acl(self.bucket_name, object_name, objectAcl)
    acl = self.client.get_object_acl(self.bucket_name, object_name)
    self.assertTrue(objectAcl.is_subset(acl))

    acl_client = GalaxyFDSClient(acl_ak, acl_access_secret,
                                 FDSClientConfiguration(region_name, False, False, False,
                                                        endpoint=endpoint))
    self.assertTrue(
      self.client.does_object_exists(self.bucket_name, object_name))
    print(acl_client.get_object(self.bucket_name, object_name))
    self.client.delete_object(self.bucket_name, object_name)
    self.assertFalse(
      self.client.does_object_exists(self.bucket_name, object_name))

  def test_get_object_and_metadata(self):
    object_name = "test1"
    content = "test1"
    self.client.put_object(self.bucket_name, object_name, content)
    whole_object = self.client.get_object(self.bucket_name, object_name)
    self.assertEqual(whole_object.get_next_chunk_as_string(), "test1")
    partial_object = self.client.get_object(self.bucket_name, object_name, 2)
    self.assertEqual(partial_object.get_next_chunk_as_string(), "st1")
    metadata = self.client.get_object_metadata(self.bucket_name, object_name)
    if IS_PY3:
      self.assertEqual(hashlib.md5("test1".encode("UTF-8")).hexdigest(),
                       metadata.metadata["content-md5"])
    else:
      self.assertEqual(hashlib.md5("test1").hexdigest(), metadata.metadata["content-md5"])
    print(metadata.metadata)

  def test_rename_object(self):
    object_old_name = "test_old1"
    object_new_name = "test_new1"
    self.client.put_object(self.bucket_name, object_old_name, "")
    self.assertTrue(
      self.client.does_object_exists(self.bucket_name, object_old_name))
    self.assertFalse(
      self.client.does_object_exists(self.bucket_name, object_new_name))
    self.client.rename_object(self.bucket_name, object_old_name, object_new_name)
    self.assertFalse(
      self.client.does_object_exists(self.bucket_name, object_old_name))
    self.assertTrue(
      self.client.does_object_exists(self.bucket_name, object_new_name))
    self.client.delete_object(self.bucket_name, object_new_name)

  def test_generate_presigned_uri(self):
    object_name = "中文测试"
    content = "presigned"
    self.client.put_object(self.bucket_name, object_name, content)
    uri = self.client.generate_presigned_uri(None, self.bucket_name, object_name,
                                             time.time() * 1000 + 60000)
    download = urlopen(uri).read()
    if IS_PY3:
      download = download.decode(encoding="UTF-8")
    self.assertEqual(content, download)

  def test_invalid_object_metadata(self):
    metadata = FDSObjectMetadata()

    metadata.add_user_metadata(FDSObjectMetadata.USER_DEFINED_METADATA_PREFIX
                               + "test", "test-value")
    metadata.add_header(Common.CACHE_CONTROL, "no-cache")

    try:
      metadata.add_user_metadata("test-key", "test-vale")
      self.fail("Exception should not be thrown here")
    except:
      pass

  def test_abort_multipart_upload(self):
    object_name = "test_object"
    upload_token = self.client.init_multipart_upload(self.bucket_name, object_name);
    self.client.abort_multipart_upload(self.bucket_name, object_name, upload_token.upload_id)

  def test_list_multi_objects(self):
    test_content = ""
    obj_prefix = "obj_"
    # add 2000 objects to make sure the result is truncated
    for i in range(2000):
      self.client.put_object(self.bucket_name, obj_prefix + str(i), test_content)
    for obj in self.client.list_all_objects(self.bucket_name):
      self.client.delete_object(self.bucket_name, obj.object_name)

  def test_multipart_upload(self):
    object_name = "test_multipart_upload"
    part_num = 3
    part_content = "1" * 5242880
    upload_list = []
    upload_token = self.client.init_multipart_upload(self.bucket_name, object_name)

    for i in range(part_num):
      upload_list.append(
        self.client.upload_part(self.bucket_name, object_name, upload_token.upload_id, i + 1,
                                part_content))

    upload_part_result = UploadPartResultList({"uploadPartResultList": upload_list})
    print(json.dumps(upload_part_result))
    self.client.complete_multipart_upload(bucket_name=self.bucket_name,
                                          object_name=object_name,
                                          upload_id=upload_token.upload_id,
                                          metadata=None,
                                          upload_part_result_list=json.dumps(upload_part_result))

    obj = self.client.get_object(self.bucket_name, object_name)
    length = 0
    for chunk in obj.stream:
      if IS_PY3:
        chunk = chunk.decode(encoding="UTF-8")
      length += len(chunk)
      for t in chunk:
        self.assertEqual(t, "1")

    obj.stream.close()
    print(length)
    self.assertEqual(length, part_num * 5242880)

  def test_delete_objects(self):
    total_count = 100;
    object_prefix = "delete-objs-";
    object_names = [""]
    for i in range(total_count):
      object_name = object_prefix + str(i)
      self.client.put_object(self.bucket_name, object_name, "")
      object_names.append(object_name)

    delete_result = self.client.delete_objects(self.bucket_name, object_names)
    self.assertEqual(len(self.client.list_objects(self.bucket_name).objects), 0)

  def test_list_trash(self):
    print("")
    total_count = 150;
    object_prefix = "trash-obj-";
    for i in range(total_count):
      self.client.put_object(self.bucket_name, object_prefix + str(i), "")

    self.client.delete_objects(self.bucket_name,
                               [object_prefix + str(i) for i in range(total_count)])

    list_result = self.client.list_trash_objects(self.bucket_name + "/trash-obj-", "", max_keys=50)
    sub_count1 = len(list_result.objects)
    self.assertEqual(50, sub_count1)
    print(list_result.next_marker)

    list_result = self.client.list_next_batch_of_objects(list_result)
    sub_count2 = len(list_result.objects)
    self.assertEqual(sub_count1 + sub_count2, total_count)

  def test_restore(self):
    print("")
    self.client.put_object(self.bucket_name, "aa/aa-1", "123")
    self.client.delete_object(self.bucket_name, "aa/aa-1")
    flag = False
    try:
      self.client.get_object(self.bucket_name, "aa/aa-1").get_next_chunk_as_string()
    except:
      flag = True

    self.assertTrue(flag)
    self.client.restore_object(self.bucket_name, "aa/aa-1")
    print(self.client.get_object(self.bucket_name, "aa/aa-1").get_next_chunk_as_string())

  def test_versioning(self):
    '''
    This test will not pass due to bucket cache in restserver
    :return:
    '''
    print("")
    versionings = [2, 3, 6, 7, 8, 12, 2]
    for v in versionings:
      self.client._update_bucket_versioning_(self.bucket_name, v)
      time.sleep(5)
      self.assertEqual(self.client._get_bucket_versioning_(self.bucket_name), v)

  def test_get_version_ids(self):
    print("")
    count = 10
    get_content = lambda i: "data-%d" % i

    for i in range(count):
      self.client.put_object(self.bucket_name, "obj", get_content(i))
    vids = self.client._list_version_ids_(self.bucket_name, "obj")
    self.assertEqual(len(vids), count - 1)
    for i, id in enumerate(reversed(vids)):
      self.assertEqual(get_content(i), self.client.get_object(self.bucket_name, "obj",
                                                              version_id=id).get_next_chunk_as_string())

  def test_lifecycle_config(self):
    print("")
    actions = [
      FDSNonCurrentVersionExpiration({"days": 0.01}),
      FDSExpiration({"days": 0.01}),
      FDSAbortIncompleteMultipartUpload({"days": 1}),
    ]

    rule1 = FDSLifecycleRule()
    rule1.enabled = True
    rule1.prefix = "image/"
    rule1.update_action(actions[0])
    rule1.update_action(actions[1])
    rule1.update_action(actions[2])

    rule2 = FDSLifecycleRule()
    rule2.enabled = True
    rule2.prefix = "image/tmp/"
    rule2.update_action(FDSExpiration({"days": 0.00001}))
    rule2.update_action(FDSNonCurrentVersionExpiration({"days": 0.00001}))

    ttlconfig = FDSLifecycleConfig()
    ttlconfig.rules.append(rule1)
    ttlconfig.rules.append(rule2)

    self.client.update_lifecycle_config(self.bucket_name, ttlconfig)
    print(json.dumps(self.client.get_lifecycle_config(self.bucket_name)))

    ttlconfig = self.client.get_lifecycle_config(self.bucket_name)
    self.assertIsNotNone(ttlconfig.get_rule_by_prefix("image/tmp/"))
    self.assertIsNotNone(ttlconfig.get_rule_by_object_name("image/tmp/123.jpg"))
    ttlconfig.get_rule_by_prefix("image/tmp/").enabled = False
    self.assertIsNotNone(ttlconfig.get_rule_by_object_name("image/tmp/123.jpg"))
    self.assertIsNotNone(
      ttlconfig.get_rule_by_object_name("image/tmp/123.jpg", enabled_rule_only=True))
    ttlconfig.get_rule_by_prefix("image/").enabled = False
    self.assertIsNotNone(ttlconfig.get_rule_by_object_name("image/tmp/123.jpg"))
    self.assertIsNone(
      ttlconfig.get_rule_by_object_name("image/tmp/123.jpg", enabled_rule_only=True))

    rule3 = FDSLifecycleRule()
    rule3.enabled = True
    rule3.prefix = "log/tmp/"
    rule3.update_action(FDSExpiration({"days": 0.001}))
    rule3.update_action(FDSNonCurrentVersionExpiration({"days": 0.001}))
    self.client.update_lifecycle_rule(self.bucket_name, rule3)
    rule = self.client.get_lifecycle_config(self.bucket_name, "3")
    self.assertEqual(rule.prefix, rule3.prefix)
    self.assertEqual(rule.prefix, rule3.prefix)
    self.assertEqual("3", rule.id)
    print(self.client.get_lifecycle_config(self.bucket_name))


  def test_cors_config(self):
    rule1=FDSCORSRule()
    rule1.id="1"
    rule1.allowOrigin='*'
    rule2=FDSCORSRule()
    rule2.id="2"
    rule2.allowOrigin='*.example.com'
    cors_config=FDSCORSConfig()
    cors_config.rules.append(rule1)
    cors_config.rules.append(rule2)
    self.client.update_cors_config(self.bucket_name,cors_config)
    print(json.dumps(self.client.get_cors_config(self.bucket_name)))
    cors_config=self.client.get_cors_config(self.bucket_name)
    self.assertIsNotNone(cors_config.get_rule_by_id("1"))
    self.assertIsNotNone(cors_config.get_rule_by_id("2"))

    rule3=FDSCORSRule()
    rule3.allowOrigin="https://cloud.d.xiaomi.net"
    self.client.update_cors_rule(self.bucket_name,rule3)
    rule = self.client.get_cors_config(self.bucket_name, "3")
    self.assertEqual("3", rule.id)
    self.assertEqual(rule.allowOrigin, rule3.allowOrigin)
    print(self.client.get_cors_config(self.bucket_name))


  def test_get_and_delete_version(self):
    print("")
    object_name = "test_get_and_delete_version"
    get_content = lambda i: "data-%d" % i

    result = self.client.put_object(self.bucket_name, object_name, get_content(0))
    self.assertIsNone(result.previous_version_id)

    count = 100
    for i in range(1, count + 1):
      result = self.client.put_object(self.bucket_name, object_name, get_content(i))
      previous_version_id = result.previous_version_id
      self.assertIsNotNone(previous_version_id)
      self.assertEqual(get_content(i - 1),
                       self.client.get_object(self.bucket_name, object_name,
                                              version_id=previous_version_id).get_next_chunk_as_string())

    vid = self.client.delete_object(self.bucket_name, object_name)
    self.assertIsNotNone(vid)
    self.assertEqual(get_content(count),
                     self.client.get_object(self.bucket_name, object_name,
                                            version_id=vid).get_next_chunk_as_string())

    vids = self.client._list_version_ids_(self.bucket_name, object_name)
    # get No.98
    self.assertEqual(get_content(98),
                     self.client.get_object(self.bucket_name, object_name,
                                            version_id=vids[2]).get_next_chunk_as_string())

  def test_auto_convert_webp(self):
    bucket_name = self.bucket_name + "-convert-webp"
    object_name = "test.jpg"

    try:
      if not self.client.does_bucket_exist(bucket_name):
        self.client.create_bucket(bucket_name)

      self.client._enable_auto_convert_webp_(bucket_name, True)
      self.assertTrue(self.client._is_enable_auto_convert_webp_(bucket_name))

      path = os.path.join(expanduser("~"), "webp-images/")
      for fn in os.listdir(path):
        if fn.endswith("jpg"):
          with open(path + fn, "rb") as f:
            self.client.put_object(bucket_name, fn, f)

      for fn in os.listdir(path):
        if fn.endswith("jpg"):
          fds_object = self.client._get_webp_(bucket_name, fn)
          with open(path + fn + ".webp", "wb") as f:
            for chunk in fds_object.stream:
              f.write(chunk)
    finally:
      ClientTest.delete_objects_and_bucket(self.client, bucket_name)




from __future__ import print_function

import json

from fds.galaxy_fds_client_exception import GalaxyFDSClientException


class FDSLifecycleConfig(dict):
  '''
  lifecycle config like this:

  {
   "rules": [
     {
       "id": 0,
       "enabled": true,
       "prefix": "log",
       "actions": {
         "nonCurrentVersionExpiration": {
           "days": 7
         },
         "expiration": {
           "days":30
         }
       }
     },
     {
       "enabled": true,
       "prefix": "images",
       "actions": {
         "nonCurrentVersionExpiration": {
           "days": 7
         },
         "expiration": {
           "days":30
         },
         "abortIncompleteMultipartUpload": {
           "days":7
         }
       }
     }
   ]
 }
  '''

  def __init__(self, json={}):
    dict.__init__(self, json)
    self._rules = []
    for rule in self.get('rules', []):
      self._rules.append(FDSLifecycleRule(rule))
    self['rules'] = self._rules

  @property
  def rules(self):
    return self._rules

  def get_rule_by_prefix(self, prefix):
    for rule in self.rules:
      if rule.prefix == prefix:
        return rule
    return None

  def get_rule_by_object_name(self, object_name, enabled_rule_only=False):
    '''
    Get rule by object name
    :param object_name:
    :param enabled_rule_only:
    :return: Only one rule will return if more than one rules matched.
    '''
    for rule in self.rules:
      if object_name.startswith(rule.prefix) and (rule.enabled or not enabled_rule_only):
        return rule
    return None


class FDSLifecycleRule(dict):
  def __init__(self, json={}):
    dict.__init__(self, json)
    self._actions = {}
    for name, action in self.get('actions', {}).items():
      if name == 'abortIncompleteMultipartUpload':
        self._actions[name] = FDSAbortIncompleteMultipartUpload(action)
      elif name == 'expiration':
        self._actions[name] = FDSExpiration(action)
      elif name == 'nonCurrentVersionExpiration':
        self._actions[name] = FDSNonCurrentVersionExpiration(action)
      elif name == 'lifeCycleStorageClass':
        self._actions[name] = FDSLifecycleStorageClass(action)
      else:
        raise GalaxyFDSClientException("invalid action type: " + name)
    self['actions'] = self._actions

  @property
  def id(self):
    return self.get('id', None)

  @property
  def enabled(self):
    return self.get('enabled', False)

  @enabled.setter
  def enabled(self, enabled):
    self['enabled'] = enabled

  @property
  def prefix(self):
    return self.get('prefix', None)

  @prefix.setter
  def prefix(self, prefix):
    self['prefix'] = prefix

  @property
  def actions(self):
    return self._actions

  def update_action(self, action):
    self.actions[action.name] = action


class FDSExpiration(dict):
  name = 'expiration'

  def __init__(self, json):
    dict.__init__(self, json)

  @property
  def days(self):
    return self.get('days', 0)

  @days.setter
  def days(self, days):
    self['days'] = days


class FDSNonCurrentVersionExpiration(dict):
  name = 'nonCurrentVersionExpiration'

  def __init__(self, json):
    dict.__init__(self, json)

  @property
  def days(self):
    return self.get('days', 0)

  @days.setter
  def days(self, days):
    self['days'] = days


class FDSAbortIncompleteMultipartUpload(dict):
  name = 'abortIncompleteMultipartUpload'

  @property
  def days(self):
    return self.get('days', 0)

  @days.setter
  def days(self, days):
    self['days'] = days

class FDSLifecycleStorageClass(dict):
  name = 'lifeCycleStorageClass'

  @property
  def days(self):
    return self.get('days', 0)

  @days.setter
  def days(self, days):
    self['days'] = days

  @property
  def storage_class(self):
    return self.get('storageClass', "")

  @storage_class.setter
  def storage_class(self, storage_class):
    self['storageClass'] = storage_class

if __name__ == '__main__':
  lifecycle_config = FDSLifecycleConfig()
  rule1 = FDSLifecycleRule()
  rule1.enabled = True
  rule1.prefix = 'test'
  action1 = FDSExpiration({'days': 30})
  rule1.update_action(action1)

  print(json.dumps(rule1, sort_keys=True))

  lifecycle_config.rules.append(rule1)

  print(json.dumps(lifecycle_config, sort_keys=True))

  lifecycle_config.rules.append(rule1)
  print(json.dumps(lifecycle_config, sort_keys=True))

  jsonstr = '''  {
   "rules": [
     {
       "enabled": true,
       "prefix": "log",
       "actions": {
         "nonCurrentVersionExpiration": {
           "days": 7
         },
         "expiration": {
           "days":30.7
         }
       }
     },
     {
       "enabled": true,
       "prefix": "images",
       "actions": {
         "nonCurrentVersionExpiration": {
           "days": 7
         },
         "expiration": {
           "days":30
         },
         "abortIncompleteMultipartUpload": {
           "days":7
         },
         "lifeCycleStorageClass": {
           "days":2,
           "storageClass":"ARCHIVE"
         }
       }
     }
   ]
 }
  '''
  print(jsonstr)
  print(json.dumps(json.loads(jsonstr), sort_keys=True))
  print(json.dumps(FDSLifecycleConfig(json.loads(jsonstr)), sort_keys=True))
